# Copyright (c) Acconeer AB, 2022
# All rights reserved

from __future__ import annotations

from typing import Callable

from acconeer.exptool.app.new import AppModel


class NullAppModel(AppModel):
    class _NullSignal:
        def connect(self, slot: Callable) -> None:
            pass

    sig_notify: _NullSignal
    sig_message_plot_plugin: _NullSignal

    def __init__(self) -> None:
        self.sig_notify = self._NullSignal()
        self.sig_message_plot_plugin = self._NullSignal()



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from __future__ import annotations

import json
from typing import Any, Tuple

import attrs
import numpy as np
import numpy.typing as npt

from .common import attrs_ndarray_eq


@attrs.frozen(kw_only=True)
class Metadata:
    """Metadata

    Represents a superset of the RSS ``processing_metadata``.
    """

    frame_data_length: int = attrs.field()
    """Number of elements in the frame"""

    sweep_data_length: int = attrs.field()
    """Number of elements in the sweep"""

    subsweep_data_offset: npt.NDArray = attrs.field(eq=attrs_ndarray_eq)
    """Offset to the subsweeps data"""

    subsweep_data_length: npt.NDArray = attrs.field(eq=attrs_ndarray_eq)
    """Number of elements in the subsweeps"""

    calibration_temperature: int = attrs.field()
    """Temperature during calibration"""

    tick_period: int = attrs.field()
    """Target tick period if update rate is set, otherwise 0"""

    base_step_length_m: float = attrs.field()
    """Base step length in meter"""

    max_sweep_rate: float = attrs.field()
    """Maximum sweep rate that the sensor can provide for the given configuration"""

    @property
    def frame_shape(self) -> Tuple[int, int]:
        """The frame shape this Metadata defines"""

        num_sweeps = self.frame_data_length // self.sweep_data_length
        return (num_sweeps, self.sweep_data_length)

    def to_dict(self) -> dict[str, Any]:
        return attrs.asdict(self)

    @classmethod
    def from_dict(cls, d: dict) -> Metadata:
        return cls(**d)

    def to_json(self) -> str:
        return json.dumps(self, cls=MetadataEncoder)

    @classmethod
    def from_json(cls, json_str: str) -> Metadata:
        return cls.from_dict(json.loads(json_str, cls=MetadataDecoder))


class MetadataEncoder(json.JSONEncoder):
    """Encoder that transforms a Metadata instance to a serializable
    dict before any json transformation
    """

    def default(self, obj: Any) -> Any:
        if isinstance(obj, Metadata):
            metadata_dict = obj.to_dict()

            # numpy arrays are not serializable.
            # Since they are integer arrays, no precision will be lost
            metadata_dict["subsweep_data_length"] = metadata_dict["subsweep_data_length"].tolist()
            metadata_dict["subsweep_data_offset"] = metadata_dict["subsweep_data_offset"].tolist()
            return metadata_dict

        return super().default(obj)


class MetadataDecoder(json.JSONDecoder):
    """Decoder that post-processes the dict (parsed from json) to better fit Metadata.from_dict"""

    def decode(self, s: str) -> Any:  # type: ignore[override]
        metadata_dict = super().decode(s)

        # post process the parsed dict to have numpy-arrays instead of plain lists
        metadata_dict["subsweep_data_length"] = np.array(metadata_dict["subsweep_data_length"])
        metadata_dict["subsweep_data_offset"] = np.array(metadata_dict["subsweep_data_offset"])

        return metadata_dict



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from __future__ import annotations

import json
from typing import Any, Optional

import attrs
import packaging.version

from acconeer.exptool.a121._core.utils import parse_rss_version


@attrs.frozen(kw_only=True)
class SensorInfo:
    """Holds information about a single sensor slot on a host."""

    connected: bool
    serial: Optional[str] = None

    def to_dict(self) -> dict[str, Any]:
        return attrs.asdict(self)

    @classmethod
    def from_dict(cls, d: dict[str, Any]) -> SensorInfo:
        return cls(**d)

    def to_json(self) -> str:
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> SensorInfo:
        return cls.from_dict(json.loads(json_str))


@attrs.frozen(kw_only=True)
class ServerInfo:
    rss_version: str = attrs.field()
    sensor_count: int = attrs.field()
    ticks_per_second: int = attrs.field()
    sensor_infos: dict[int, SensorInfo] = attrs.field()
    hardware_name: Optional[str] = attrs.field(default=None)

    @property
    def parsed_rss_version(self) -> packaging.version.Version:
        return parse_rss_version(self.rss_version)

    def to_dict(self) -> dict[str, Any]:
        return attrs.asdict(self)

    @classmethod
    def from_dict(cls, d: dict) -> ServerInfo:
        d = d.copy()
        d["sensor_infos"] = {
            int(sensor_id): SensorInfo.from_dict(sensor_info_dict)
            for sensor_id, sensor_info_dict in d["sensor_infos"].items()
        }
        return cls(**d)

    def to_json(self) -> str:
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> ServerInfo:
        return cls.from_dict(json.loads(json_str))



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from collections import namedtuple

import numpy as np

from acconeer.exptool.a111._modes import Mode, get_mode


RegVal = namedtuple("RegVal", ["addr", "val"])
RegReadRequest = namedtuple("RegReadRequest", ["addr"])
RegWriteRequest = namedtuple("RegWriteRequest", ["reg_val"])
RegReadResponse = namedtuple("RegReadResponse", ["reg_val"])
RegWriteResponse = namedtuple("RegWriteResponse", ["reg_val"])
BufferReadRequest = namedtuple("BufferReadRequest", ["addr"])
BufferReadResponse = namedtuple("BufferReadResponse", ["addr", "buffer"])
StreamData = namedtuple("StreamData", ["result_info", "buffer"])


class ProtocolError(Exception):
    pass


ADDR_SIZE = 1
REG_SIZE = 4
LEN_FIELD_SIZE = 2
MIN_FRAME_SIZE = 1 + LEN_FIELD_SIZE + 1 + 1
BYTEORDER = "little"
BO = BYTEORDER

START_MARKER = 0xCC
END_MARKER = 0xCD
REG_READ_REQUEST = 0xF8
REG_READ_RESPONSE = 0xF6
REG_WRITE_REQUEST = 0xF9
REG_WRITE_RESPONSE = 0xF5
STREAM_PACKET = 0xFE
BUF_READ_REQUEST = 0xFA
BUF_READ_RESPONSE = 0xF7
STREAM_RESULT_INFO = 0xFD
STREAM_BUFFER = 0xFE

MAIN_BUFFER_ADDR = 0xE8


def unpack_packet(packet):
    if len(packet) < 1:
        raise ProtocolError("package is too short")

    packet_type = packet[0]
    segment = packet[1:]

    if packet_type == REG_READ_RESPONSE:
        return unpack_reg_read_res_segment(segment)
    elif packet_type == REG_WRITE_RESPONSE:
        return unpack_reg_write_res_segment(segment)
    elif packet_type == BUF_READ_RESPONSE:
        return unpack_buf_read_res_segment(segment)
    elif packet_type == STREAM_PACKET:
        return unpack_stream_data_segment(segment)
    else:
        raise ProtocolError("unknown packet type")


def unpack_reg_val(packed):
    if len(packed) != ADDR_SIZE + REG_SIZE:
        raise ProtocolError("unexpected package length")

    reg_addr = packed[0]
    enc_val = packed[1:]

    return RegVal(reg_addr, enc_val)


def unpack_reg_read_res_segment(segment):
    rv = unpack_reg_val(segment)
    return RegReadResponse(rv)


def unpack_reg_write_res_segment(segment):
    rv = unpack_reg_val(segment)
    return RegWriteResponse(rv)


def unpack_buf_read_res_segment(segment):
    buf_addr = segment[0]
    buffer = segment[1:]
    return BufferReadResponse(buf_addr, buffer)


def unpack_stream_data_segment(segment):
    result_info = None
    buffer = None
    rest = segment
    while len(rest) > 0:
        if len(rest) < 1 + LEN_FIELD_SIZE:
            raise ProtocolError("invalid package length")

        part_type = rest[0]
        data_start_index = 1 + LEN_FIELD_SIZE
        part_len = int.from_bytes(rest[1:data_start_index], BO)
        data_end_index = data_start_index + part_len
        part_data = rest[data_start_index:data_end_index]
        rest = rest[data_end_index:]

        if part_type == STREAM_RESULT_INFO:
            s = ADDR_SIZE + REG_SIZE
            if part_len % s != 0:
                raise ProtocolError("invalid package length")

            result_info = []
            num_regs = part_len // s
            for i in range(num_regs):
                addr = part_data[s * i]
                enc_val = part_data[s * i + 1 : s * (i + 1)]
                rrv = RegVal(addr, enc_val)
                result_info.append(rrv)
        elif part_type == STREAM_BUFFER:
            buffer = part_data
        else:
            raise ProtocolError("unknown stream part type")

    return StreamData(result_info, buffer)


def pack_reg_val(reg_val):
    if len(reg_val.val) != REG_SIZE:
        raise ProtocolError("register value must be {} bytes".format(REG_SIZE))
    packed = bytearray()
    packed.extend(reg_val.addr.to_bytes(ADDR_SIZE, BO))
    packed.extend(reg_val.val)
    return packed


def pack_packet(packet):
    if isinstance(packet, RegReadRequest):
        packet_type = REG_READ_REQUEST
        packet_data = bytearray()
        packet_data.extend(packet.addr.to_bytes(ADDR_SIZE, BO))
    elif isinstance(packet, RegWriteRequest):
        packet_type = REG_WRITE_REQUEST
        packet_data = pack_reg_val(packet.reg_val)
    elif isinstance(packet, RegReadResponse):
        packet_type = REG_READ_RESPONSE
        packet_data = pack_reg_val(packet.reg_val)
    elif isinstance(packet, RegWriteResponse):
        packet_type = REG_WRITE_RESPONSE
        packet_data = pack_reg_val(packet.reg_val)
    elif isinstance(packet, BufferReadRequest):
        packet_type = BUF_READ_REQUEST
        packet_data = bytearray()
        packet_data.extend(packet.addr.to_bytes(ADDR_SIZE, BO))
        packet_data.extend([0, 0])
    else:
        raise TypeError("unknown type of packet")

    packet_bytes = bytearray()
    packet_bytes.append(packet_type)
    packet_bytes.extend(packet_data)
    return packet_bytes


def extract_packet_from_frame(frame):
    if len(frame) < MIN_FRAME_SIZE:
        raise ProtocolError("invalid frame (frame too short)")
    if frame[0] != START_MARKER:
        raise ProtocolError("invalid frame (incorrect start marker)")
    if frame[-1] != END_MARKER:
        raise ProtocolError("invalid frame (incorrect start marker)")

    packet_len = int.from_bytes(frame[1 : 1 + LEN_FIELD_SIZE], BO)
    packet = frame[1 + LEN_FIELD_SIZE : -1]

    if len(packet) - 1 != packet_len:
        raise ProtocolError("invalid frame (packet length mismatch)")

    return packet


def insert_packet_into_frame(packet):
    if not isinstance(packet, bytearray):
        packet = pack_packet(packet)

    packet_len = len(packet) - 1
    frame = bytearray()
    frame.append(START_MARKER)
    frame.extend(packet_len.to_bytes(LEN_FIELD_SIZE, BO))
    frame.extend(packet)
    frame.append(END_MARKER)
    return frame


def decode_output_buffer(buffer, mode, sweeps_per_frame=None):
    mode = get_mode(mode)

    if mode == Mode.POWER_BINS:
        return np.frombuffer(buffer, dtype="<u2").astype("float")
    elif mode == Mode.ENVELOPE:
        return np.frombuffer(buffer, dtype="<u2").astype("float")
    elif mode == Mode.IQ:
        data = np.frombuffer(buffer, dtype="<i2").astype("float")
        return data.reshape((-1, 2)).view(dtype="complex").flatten()
    elif mode == Mode.SPARSE:
        data = np.frombuffer(buffer, dtype="<u2").astype("float")
        data = data.reshape((sweeps_per_frame, -1))
        return data
    else:
        raise NotImplementedError



# Copyright (c) Acconeer AB, 2022
# All rights reserved

import logging
from time import sleep, time

import numpy as np

from acconeer.exptool.a111 import SDK_VERSION
from acconeer.exptool.a111._clients.base import BaseClient, ClientError, decode_version_str
from acconeer.exptool.a111._configs import BaseServiceConfig
from acconeer.exptool.a111._modes import Mode


log = logging.getLogger(__name__)


START_KEY = "range_start_m"
LENGTH_KEY = "range_length_m"
STEP_LENGTH_KEY = "step_length_m"
MISSED_GET_NEXT_KEY = "missed_data"
DATA_SATURATED_KEY = "data_saturated"
DATA_QUALITY_WARNING_KEY = "data_quality_warning"
DATA_LENGTH_KEY = "data_length"


class MockClient(BaseClient):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def _connect(self):
        info = {}
        info.update(decode_version_str(SDK_VERSION))
        info["mock"] = True
        return info

    def _setup_session(self, config):
        self._config = config

        update_rate_limit = 100

        if config.mode == Mode.SPARSE:
            if config.sweep_rate is not None:
                sparse_frame_rate_limit = config.sweep_rate / config.sweeps_per_frame
                update_rate_limit = min(update_rate_limit, sparse_frame_rate_limit)

        if config.update_rate is None:
            self._update_rate = update_rate_limit
            self._missed = False
        else:
            self._update_rate = min(config.update_rate, update_rate_limit)
            self._missed = config.update_rate > self._update_rate

        try:
            mock_class = MOCK_CLASS_MAP[config.mode]
        except KeyError as e:
            raise ClientError("mode not supported") from e

        self._mocker = mock_class(config)
        info = self._mocker.session_info
        info["stitch_count"] = 0
        return info

    def _start_session(self):
        self._start_time = time()
        self._data_count = 0

    def _get_next(self):
        config = self._config

        self._data_count += 1

        data_capture_time = self._data_count / self._update_rate
        now = time() - self._start_time
        if data_capture_time > now:
            sleep(data_capture_time - now)

        args = (data_capture_time, self._data_count)
        num_sensors = len(config.sensor)

        if self.squeeze and num_sensors == 1:
            info, data = self._mocker.get_next(*args, 0)
            info[MISSED_GET_NEXT_KEY] = self._missed
        else:
            idx_offset = max(0, (num_sensors - 1) / 2)
            out = [self._mocker.get_next(*args, i - idx_offset) for i in range(num_sensors)]
            info, data = zip(*out)
            data = np.array(data)
            info = list(info)

            for d in info:
                d[MISSED_GET_NEXT_KEY] = self._missed

        return info, data

    def _stop_session(self):
        pass

    def _disconnect(self):
        pass

    @property
    def description(self):
        return "simulated interface"


class DenseMocker:
    BASE_STEP_LENGTH = 0.485e-3

    def __init__(self, config):
        self.config = config

        step_length = self.BASE_STEP_LENGTH * config.downsampling_factor

        self.num_depths = int(round(config.range_length / step_length)) + 1

        self.session_info = {
            START_KEY: config.range_start,
            LENGTH_KEY: config.range_length,
            DATA_LENGTH_KEY: self.num_depths,
            STEP_LENGTH_KEY: step_length,
        }

        self.range_center = config.range_start + config.range_length / 2
        self.depths = np.linspace(*config.range_interval, self.num_depths)


class EnvelopeMocker(DenseMocker):
    def get_next(self, t, i, offset):
        info = {
            DATA_SATURATED_KEY: False,
            DATA_QUALITY_WARNING_KEY: False,
        }

        noise = 100 + 20 * np.random.randn(self.num_depths)
        noise = filtfilt_simple(noise, 0.98)

        ampl = 2000 + np.random.randn() * 20
        center = self.range_center
        center += np.random.randn() * 0.2e-3
        center += offset * 0.1
        profile = getattr(self.config, "profile", BaseServiceConfig.Profile.PROFILE_2)
        s = 0.01 + (profile.json_value - 1.0) * 0.03
        signal = ampl * np.exp(-np.square((self.depths - center) / s))

        data = signal + noise

        data = np.rint(data)

        return info, data


class IQMocker(DenseMocker):
    def get_next(self, t, i, offset):
        info = {
            DATA_SATURATED_KEY: False,
            DATA_QUALITY_WARNING_KEY: False,
        }

        noise = np.random.randn(self.num_depths) + 1j * np.random.randn(self.num_depths)
        noise *= 0.015

        ampl = 0.2 * (1 + 0.03 * np.random.randn())
        center = self.range_center
        center += np.random.randn() * (3 / 360) * 2.5e-3
        center += offset * 0.1
        center += 4e-3 * np.sin(t)
        xs = self.depths - center
        signal = ampl * np.exp(2j * np.pi * xs / 2.5e-3) * np.exp(-np.square(xs / 0.05))

        data = signal + noise
        data *= np.exp(-2j * np.pi * self.depths / 2.5e-3)
        data = filtfilt_simple(data, 0.98)

        return info, data


class PowerBinMocker(EnvelopeMocker):
    def __init__(self, config):
        self.config = config

        step_length = self.BASE_STEP_LENGTH * config.downsampling_factor

        self.num_depths = config.bin_count or int(round(config.range_length / 0.1)) + 1

        self.session_info = {
            START_KEY: config.range_start,
            LENGTH_KEY: config.range_length,
            DATA_LENGTH_KEY: self.num_depths,
            STEP_LENGTH_KEY: step_length,
            "bin_count": self.num_depths,
        }

        self.range_center = config.range_start + config.range_length / 2
        self.depths = np.linspace(*config.range_interval, self.num_depths)


class SparseMocker:
    BASE_STEP_LENGTH = 0.06

    def __init__(self, config):
        self.config = config

        step_length = 0.06 * config.downsampling_factor

        start_point = int(round(config.range_start / self.BASE_STEP_LENGTH))
        start = start_point * self.BASE_STEP_LENGTH
        length_point = int(round((config.range_end - start) / step_length))
        self.num_depths = length_point + 1
        end_point = start_point + length_point
        end = end_point * self.BASE_STEP_LENGTH

        self.session_info = {
            START_KEY: start,
            LENGTH_KEY: end - start,
            DATA_LENGTH_KEY: self.num_depths * config.sweeps_per_frame,
            STEP_LENGTH_KEY: step_length,
            "sweep_rate": config.sweep_rate or 5e3,
        }

        self.range_center = (start + end) / 2
        self.depths = np.linspace(start, end, self.num_depths)

    def get_next(self, t, i, offset):
        info = {
            DATA_SATURATED_KEY: False,
        }

        num_sweeps = self.config.sweeps_per_frame

        noise = 100 * np.random.randn(num_sweeps, self.num_depths)

        xs = self.depths - self.range_center + 0.1 * np.sin(t)
        signal = 5000 * np.exp(-np.square(xs / 0.1)) * np.sin(xs / 2.5e-3)

        data = 2**15 + noise + np.tile(signal[None, :], [num_sweeps, 1])

        data = np.rint(data)

        return info, data


def lfilter_simple(x, sf):
    y = np.zeros_like(x)
    y[0] = x[0]

    for i in range(1, len(x)):
        y[i] = sf * y[i - 1] + (1 - sf) * x[i]

    return y


def filtfilt_simple(x, sf):
    return np.flip(lfilter_simple(np.flip(lfilter_simple(x, sf)), sf))


MOCK_CLASS_MAP = {
    Mode.ENVELOPE: EnvelopeMocker,
    Mode.IQ: IQMocker,
    Mode.SPARSE: SparseMocker,
    Mode.POWER_BINS: PowerBinMocker,
}



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from __future__ import annotations

import abc
import subprocess
from pathlib import Path
from typing import List


class SetupStep(abc.ABC):
    @abc.abstractmethod
    def report(self) -> None:
        """Prints a summary of what this setup step will do.

        This method will be called before `install()` to let the user
        abort.
        """
        pass

    @abc.abstractmethod
    def run(self) -> bool:
        """Performs the setup steps

        :returns: if the step ran successfully.
        """
        pass


class ShellCommandStep(SetupStep):
    """Runs a shell command as a step"""

    def __init__(self, command: List[str]) -> None:
        self.command = command

    def __str__(self) -> str:
        return "$ " + (" ".join(self.command))

    def report(self) -> None:
        print(self)

    def run(self) -> bool:
        print(f">>> {self!s}")
        completed_process = subprocess.run(self.command)
        return completed_process.returncode == 0


class RequireFileContentStep(ShellCommandStep):
    """Makes sure `required_content` is the exact content in `file_path`.

    If `file_path` does not exist, an attempt will be made to create it with `required_countent`.
    """

    def __init__(self, file_path: str, required_content: str, sudo: bool = False) -> None:
        super().__init__(
            (["sudo"] if sudo else [])
            + ["sh", "-c", f"echo -n '{required_content}' > {file_path}"]
        )
        self.file_path = Path(file_path)
        self.required_content = required_content

    def report(self) -> None:
        print(f"Will make sure that {self.file_path} contains this content:")
        print()
        print(f"    {self.required_content!r}")
        print()

    def run(self) -> bool:
        print(f">>> checking {self.file_path}")

        if self.file_path.exists():
            return self._report_on_content()

        creation_ok = super().run()
        return creation_ok and self._report_on_content()

    def _report_on_content(self) -> bool:
        if self.required_content == self.file_path.read_text():
            print(f"{self.file_path} looks good.")
            return True

        print(f"The contents of {self.file_path} does not match.")
        print()
        print("Expected:")
        print(f"    {self.required_content}")
        print()
        print("Actual:")
        print(f"    {self.file_path.read_text()}")
        return False



# Copyright (c) Acconeer AB, 2022
# All rights reserved

"""
This file handles the lower layer of the UART protocol.
"""
import logging
import queue
import threading
from enum import Enum


_LOG = logging.getLogger(__name__)


class Packet:
    """
    Base class for all packages transmitted over UART protocol.
    """

    PROTOCOL_START_MARKER = 0xCC
    PROTOCOL_END_MARKER = 0xCD

    def __init__(self, payload):
        self.payload = payload

    @property
    def packet_type(self):
        raise NotImplementedError

    def get_byte_array(self):
        """
        Packet the data for this packet into a bytearray()
        that can be transmitted over UART.
        """
        data = bytearray([])
        data.append(Packet.PROTOCOL_START_MARKER)
        packet_length = int(len(self.payload))
        data.extend(packet_length.to_bytes(2, byteorder="little"))
        data.append(self.packet_type)
        data.extend(self.payload)
        data.append(Packet.PROTOCOL_END_MARKER)
        return data


class UartReader(threading.Thread):
    """Class that reads packets from the serial port"""

    HEADER_LENGTH = 3  # Excluding start marker
    PACKET_TIMEOUT = 0.5
    DEFAULT_READ_TIMEOUT = 10
    DEFAULT_READ_STREAM_TIMEOUT = 2
    packet_count = 0

    def __init__(self, ser, response_classes):
        """
        Create a reader that reads packages from the UART.

        ser: The serial port object to read from
        packet_types: An array with response package classes
        """
        threading.Thread.__init__(self, name="UartReader")
        self._ser = ser
        self._stop_event = threading.Event()
        self.daemon = True
        self._packets = dict()
        self._packet_types = dict()
        # Create a map with key=packet_type, value=packet class
        # also create a queue for each possible packet type
        for response_class in response_classes:
            if response_class.packet_type in self._packet_types:
                raise ValueError(f"Duplicate package type specified: {response_class.packet_type}")
            if not isinstance(response_class.packet_type, int):
                raise ValueError(
                    f"packet_type must be an int: {response_class}: {response_class.packet_type}"
                )
            self._packet_types[response_class.packet_type] = response_class
            self._packets[response_class.packet_type] = queue.Queue()

    def stop(self):
        self._stop_event.set()
        self._ser.cancel_read()
        self.join()

    def _wait_for_start_marker(self):
        self._ser.timeout = None
        errors = 0
        while not self._stop_event.is_set():
            data = self._ser.read(1)
            if len(data) != 1:
                continue

            if data[0] != Packet.PROTOCOL_START_MARKER:
                if errors == 0:
                    # Invalid start marker before first packet is not an error
                    # but rather some data that was around from previous session
                    _LOG.log(
                        logging.WARNING if self.packet_count == 0 else logging.ERROR,
                        "Invalid start marker %d",
                        data[0],
                    )
                errors += 1
            else:
                break

        if errors != 0:
            _LOG.log(
                logging.WARNING if self.packet_count == 0 else logging.ERROR,
                "Got %d invalid start markers",
                errors,
            )

        self.packet_count += 1

    def _read_packet_header(self):
        _LOG.debug("Reading packet header")
        self._ser.timeout = UartReader.PACKET_TIMEOUT

        header = self._ser.read(UartReader.HEADER_LENGTH)
        if self._stop_event.is_set():
            return (None, None)

        if len(header) != UartReader.HEADER_LENGTH:
            _LOG.error("Invalid header length %d", len(header))
            return (None, None)

        _LOG.debug("Got packet header %s", header)

        payload_length = header[0] | header[1] << 8
        packet_type = header[2]
        if packet_type not in self._packet_types:
            _LOG.error("Unknown packet type 0x%02X", packet_type)
            return (None, None)

        return (payload_length, packet_type)

    def _read_payload_and_end_marker(self, payload_length):
        _LOG.debug("Reading payload and end marker len=%d", payload_length)
        self._ser.timeout = (
            UartReader.PACKET_TIMEOUT + (payload_length + 1) / self._ser.baudrate * 10
        )
        payload = self._ser.read(payload_length + 1)
        if payload and len(payload) == (payload_length + 1):
            return (payload[:payload_length], payload[payload_length])
        _LOG.debug("wrong payload length %d was %s", len(payload), payload)
        return (None, None)

    def run(self):
        _LOG.debug("UartReader starting")
        try:
            while not self._stop_event.is_set():
                self._wait_for_start_marker()
                if self._stop_event.is_set():
                    break

                (payload_length, packet_type) = self._read_packet_header()
                if self._stop_event.is_set():
                    break

                if not payload_length:
                    continue

                (payload, end_marker) = self._read_payload_and_end_marker(payload_length)

                if self._stop_event.is_set():
                    break

                if not payload:
                    _LOG.error("Protocol out of sync, starting over (payload)")
                    continue
                if end_marker != Packet.PROTOCOL_END_MARKER:
                    _LOG.error("Invalid end marker: %s", end_marker)
                    continue

                _LOG.debug(
                    "Got packet type=%s payload[:10]=%s end_marker=%s",
                    packet_type,
                    payload[:10],
                    end_marker,
                )
                packet = self._packet_types[packet_type](payload)
                self._packets[packet_type].put(packet)
        except Exception as exception:
            _LOG.error("UartReader: Got exception %s", exception)
            raise exception
        finally:
            _LOG.debug("UartReader stopping")

    def have_packet(self, packet_type):
        key = packet_type.value if isinstance(packet_type, Enum) else packet_type
        return not self._packets[key].empty()

    def wait_packet(self, packet_type, timeout):
        _LOG.debug("Waiting for packet of type %s with a timeout of %.2f", packet_type, timeout)
        key = packet_type.value if isinstance(packet_type, Enum) else packet_type
        packet = self._packets[key].get(timeout=timeout)
        packet_type_value = packet_type.value if isinstance(packet_type, Enum) else packet_type
        _LOG.debug("Got packet of type %s (0x%02X)", packet_type, packet_type_value)
        return packet



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from .entities import (
    INT_16_COMPLEX,
    PRF,
    ClientInfo,
    Criticality,
    IdleState,
    Metadata,
    PersistentRecord,
    Profile,
    Record,
    Result,
    SensorConfig,
    SensorInfo,
    ServerInfo,
    SessionConfig,
    StackedResults,
    SubsweepConfig,
    ValidationError,
    ValidationResult,
    ValidationWarning,
)
from .mediators import ClientError, Recorder
from .peripherals import (
    _H5PY_STR_DTYPE,
    Client,
    H5Record,
    H5Recorder,
    InMemoryRecord,
    ServerError,
    load_record,
    open_record,
    save_record,
    save_record_to_h5,
)
from .utils import iterate_extended_structure



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from __future__ import annotations

from argparse import SUPPRESS, ArgumentParser, Namespace
from typing import Any, Dict

import numpy as np

import acconeer.exptool as et


def get_range_depths(sensor_config: et.configbase.SensorConfig, session_info: dict) -> np.ndarray:
    """Get range depths in meters."""

    range_start = session_info["range_start_m"]
    range_end = range_start + session_info["range_length_m"]

    if sensor_config.mode == et.a111.Mode.SPARSE:
        num_depths = session_info["data_length"] // sensor_config.sweeps_per_frame
    elif sensor_config.mode == et.a111.Mode.POWER_BINS:
        num_depths = session_info["bin_count"]
    else:
        num_depths = session_info["data_length"]

    return np.linspace(range_start, range_end, num_depths)


class ExampleArgumentParser(ArgumentParser):
    def __init__(self, num_sens="+"):
        super().__init__()

        server_group = self.add_mutually_exclusive_group(required=True)
        server_group.add_argument(
            "-u",
            "--uart",
            metavar="port",
            dest="serial_port",
            help="connect via uart (using module protocol by default)",
            nargs="?",
            const="",  # as argparse does not support setting const to None
        )
        server_group.add_argument(
            "-s",
            "--socket",
            metavar="address",
            dest="socket_addr",
            help="connect via socket on given address (using json-based protocol)",
        )
        server_group.add_argument(
            "-spi",
            "--spi",
            dest="spi",
            help="connect via spi (using register-based protocol)",
            action="store_true",
        )

        self.add_argument(
            "--protocol",
            metavar="protocol",
            help='What specific protocol to use. Any of "module", "exploration" or "streaming"',
            choices=["module", "exploration", "streaming"],
            default=SUPPRESS,
        )

        self.add_argument(
            "--sensor",
            metavar="id",
            dest="sensors",
            type=int,
            default=[1],
            nargs=num_sens,
            help="the sensor(s) to use (default: 1)",
        )
        verbosity_group = self.add_mutually_exclusive_group(required=False)
        verbosity_group.add_argument(
            "-v",
            "--verbose",
            action="store_true",
        )
        verbosity_group.add_argument(
            "-vv",
            "--debug",
            action="store_true",
        )
        verbosity_group.add_argument(
            "-q",
            "--quiet",
            action="store_true",
        )


def get_client_args(namespace: Namespace) -> Dict[str, Any]:
    """
    Filters a passed Namespace to extract client-creation related args.

    :returns: dictionary with client-related keyword-arguments.
    """
    result = {}
    ns_dict = {k: v for k, v in vars(namespace).items() if v is not None}

    if "protocol" in ns_dict:
        result["protocol"] = ns_dict["protocol"]

    if "socket_addr" in ns_dict:
        result["host"] = ns_dict["socket_addr"]
        result["link"] = et.a111.Link.SOCKET
        return result

    if "spi" in ns_dict and ns_dict["spi"]:
        result["link"] = et.a111.Link.SPI
        return result

    if "serial_port" in ns_dict:
        if ns_dict["serial_port"] != "":
            result["serial_port"] = ns_dict["serial_port"]
        if "protocol" not in result:
            result["protocol"] = et.a111.Protocol.MODULE
        result["link"] = et.a111.Link.UART
        return result

    return result



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from __future__ import annotations

import copy
import enum
import warnings
from typing import Any, Dict, List, Optional, Tuple

import attrs
import h5py
import numpy as np
import numpy.typing as npt

from acconeer.exptool import a121
from acconeer.exptool.a121.algo import AlgoConfigBase

from ._aggregator import (
    Aggregator,
    AggregatorConfig,
    AggregatorContext,
    PeakSortingMethod,
    ProcessorSpec,
)
from ._processors import (
    DEFAULT_CFAR_ONE_SIDED,
    DEFAULT_FIXED_THRESHOLD_VALUE,
    DEFAULT_THRESHOLD_SENSITIVITY,
    MeasurementType,
    Processor,
    ProcessorConfig,
    ProcessorContext,
    ProcessorMode,
    ProcessorResult,
    ThresholdMethod,
    calculate_bg_noise_std,
    calculate_offset,
)


@attrs.frozen(kw_only=True)
class DetectorStatus:
    detector_state: DetailedStatus
    ready_to_calibrate_close_range: bool
    ready_to_record_threshold: bool
    ready_to_start: bool


class DetailedStatus(enum.Enum):
    OK = enum.auto()
    CLOSE_RANGE_CALIBRATION_MISSING = enum.auto()
    CLOSE_RANGE_CALIBRATION_CONFIG_MISMATCH = enum.auto()
    RECORDED_THRESHOLD_MISSING = enum.auto()
    RECORDED_THRESHOLD_CONFIG_MISMATCH = enum.auto()
    INVALID_DETECTOR_CONFIG_RANGE = enum.auto()


@attrs.frozen(kw_only=True)
class SubsweepGroupPlan:
    step_length: int = attrs.field()
    breakpoints: list[int] = attrs.field()
    profile: a121.Profile = attrs.field()
    hwaas: list[int] = attrs.field()


Plan = Dict[MeasurementType, List[SubsweepGroupPlan]]


@attrs.mutable(kw_only=True)
class DetectorContext:
    offset_m: Optional[float] = attrs.field(default=None)
    direct_leakage: Optional[npt.NDArray[np.complex_]] = attrs.field(default=None)
    phase_jitter_comp_reference: Optional[npt.NDArray[np.float_]] = attrs.field(default=None)
    recorded_thresholds_mean_sweep: Optional[List[npt.NDArray[np.float_]]] = attrs.field(
        default=None
    )
    recorded_thresholds_noise_std: Optional[List[List[np.float_]]] = attrs.field(default=None)
    bg_noise_std: Optional[List[List[float]]] = attrs.field(default=None)
    recorded_threshold_session_config_used: Optional[a121.SessionConfig] = attrs.field(
        default=None
    )
    close_range_session_config_used: Optional[a121.SessionConfig] = attrs.field(default=None)

    # TODO: Make recorded_thresholds Optional[List[Optional[npt.NDArray[np.float_]]]]


@attrs.mutable(kw_only=True)
class DetectorConfig(AlgoConfigBase):
    start_m: float = attrs.field(default=0.2)
    """Start point of measurement interval in meters."""

    end_m: float = attrs.field(default=1.0)
    """End point of measurement interval in meters."""

    max_step_length: Optional[int] = attrs.field(default=None)  # TODO: Check validity
    """Used to limit step length. If no argument is provided, the step length is automatically
    calculated based on the profile."""

    max_profile: a121.Profile = attrs.field(default=a121.Profile.PROFILE_5, converter=a121.Profile)
    """Specifies the longest allowed profile. If no argument is provided, the highest possible
    profile without interference of direct leakage is used to maximize SNR."""

    signal_quality: float = attrs.field(default=15.0)
    """Signal quality. High quality equals higher HWAAS and better SNR but increase power
    consumption."""

    threshold_method: ThresholdMethod = attrs.field(
        default=ThresholdMethod.CFAR,
        converter=ThresholdMethod,
    )
    """Threshold method"""

    peaksorting_method: PeakSortingMethod = attrs.field(
        default=PeakSortingMethod.STRONGEST,
        converter=PeakSortingMethod,
    )
    """Sorting method of estimated distances."""

    num_frames_in_recorded_threshold: int = attrs.field(default=100)
    """Number of frames used when calibrating threshold."""

    fixed_threshold_value: float = attrs.field(default=DEFAULT_FIXED_THRESHOLD_VALUE)
    """Value of fixed threshold."""

    threshold_sensitivity: float = attrs.field(default=DEFAULT_THRESHOLD_SENSITIVITY)
    """Sensitivity of threshold. High sensitivity equals low detection threshold, low sensitivity
    equals high detection threshold."""

    cfar_one_sided: bool = attrs.field(default=DEFAULT_CFAR_ONE_SIDED)
    """Use one sided CFAR threshold."""


@attrs.frozen(kw_only=True)
class DetectorResult:
    distances: Optional[npt.NDArray[np.float_]] = attrs.field(default=None)
    processor_results: list[ProcessorResult] = attrs.field()
    service_extended_result: list[dict[int, a121.Result]] = attrs.field()


class Detector:
    """Distance detector
    :param client: Client
    :param sensor_id: Sensor id
    :param detector_config: Detector configuration
    :param context: Detector context
    """

    MIN_DIST_M = 0.0
    MAX_DIST_M = 17.0
    MAX_MEASURABLE_DIST_M = {
        a121.PRF.PRF_19_5_MHz: 3.1,
        a121.PRF.PRF_13_0_MHz: 7.0,
        a121.PRF.PRF_8_7_MHz: 12.7,
        a121.PRF.PRF_6_5_MHz: 18.5,
    }
    MIN_LEAKAGE_FREE_DIST_M = {
        a121.Profile.PROFILE_1: 0.12,
        a121.Profile.PROFILE_2: 0.28,
        a121.Profile.PROFILE_3: 0.56,
        a121.Profile.PROFILE_4: 0.76,
        a121.Profile.PROFILE_5: 1.28,
    }
    MIN_NUM_POINTS_IN_ENVELOPE_FWHM_SPAN = 4.0
    VALID_STEP_LENGTHS_IN_COARSE = [1, 2, 3, 4, 6, 8, 12, 24]
    NUM_POINTS_IN_COARSE = 24
    NUM_SUBSWEEPS_IN_SENSOR_CONFIG = 4

    MAX_HWAAS = 511
    MIN_HWAAS = 1
    HWAAS_MIN_DISTANCE = 1.0
    RLG_PER_HWAAS_MAP = {
        a121.Profile.PROFILE_1: 11.3,
        a121.Profile.PROFILE_2: 13.7,
        a121.Profile.PROFILE_3: 19.0,
        a121.Profile.PROFILE_4: 20.5,
        a121.Profile.PROFILE_5: 21.6,
    }

    session_config: a121.SessionConfig
    processor_specs: List[ProcessorSpec]
    context: DetectorContext

    def __init__(
        self,
        *,
        client: a121.Client,
        sensor_id: int,
        detector_config: DetectorConfig,
        context: Optional[DetectorContext] = None,
    ) -> None:
        self.client = client
        self.sensor_id = sensor_id
        self.detector_config = detector_config
        self.started = False

        if context is None:
            self.context = DetectorContext()
        else:
            self.context = context

        self.aggregator: Optional[Aggregator] = None

        self.update_config(self.detector_config)

    def _validate_ready_for_calibration(self) -> None:
        if self.started:
            raise RuntimeError("Already started")
        if self.processor_specs is None:
            raise ValueError("Processor specification not defined")
        if self.session_config is None:
            raise ValueError("Session config not defined")

    def calibrate_close_range(self) -> None:
        self._validate_ready_for_calibration()

        close_range_spec = self._filter_close_range_spec(self.processor_specs)
        spec = self._update_processor_mode(close_range_spec, ProcessorMode.LEAKAGE_CALIBRATION)

        # Note - Setup with full session_config to match the structure of spec
        extended_metadata = self.client.setup_session(self.session_config)
        assert isinstance(extended_metadata, list)

        aggregator = Aggregator(
            session_config=self.session_config,
            extended_metadata=extended_metadata,
            config=AggregatorConfig(),
            context=AggregatorContext(),
            specs=spec,
        )

        self.client.start_session()
        extended_result = self.client.get_next()
        assert isinstance(extended_result, list)
        aggregator_result = aggregator.process(extended_result=extended_result)
        self.client.stop_session()
        (processor_result,) = aggregator_result.processor_results
        self.context.direct_leakage = processor_result.direct_leakage
        assert processor_result.phase_jitter_comp_reference is not None
        self.context.phase_jitter_comp_reference = processor_result.phase_jitter_comp_reference
        self.context.close_range_session_config_used = self.session_config
        self.context.recorded_thresholds_mean_sweep = None
        self.context.recorded_thresholds_noise_std = None

    def record_threshold(self) -> None:
        self._validate_ready_for_calibration()

        # TODO: Ignore/override threshold method while recording threshold

        specs_updated = self._update_processor_mode(
            self.processor_specs, ProcessorMode.RECORDED_THRESHOLD_CALIBRATION
        )

        self.calibrate_noise()

        specs = self._add_context_to_processor_spec(specs_updated)

        extended_metadata = self.client.setup_session(self.session_config)
        assert isinstance(extended_metadata, list)

        aggregator = Aggregator(
            session_config=self.session_config,
            extended_metadata=extended_metadata,
            config=AggregatorConfig(),
            context=AggregatorContext(),
            specs=specs,
        )

        aggregator_result = None
        self.client.start_session()
        for _ in range(self.detector_config.num_frames_in_recorded_threshold):
            extended_result = self.client.get_next()
            assert isinstance(extended_result, list)
            aggregator_result = aggregator.process(extended_result=extended_result)
        self.client.stop_session()

        assert aggregator_result is not None

        self.context.recorded_thresholds_mean_sweep = []
        self.context.recorded_thresholds_noise_std = []
        for processor_result in aggregator_result.processor_results:
            # Since we know what mode the processor is running in
            assert processor_result.recorded_threshold_mean_sweep is not None
            assert processor_result.recorded_threshold_noise_std is not None
            self.context.recorded_thresholds_mean_sweep.append(
                processor_result.recorded_threshold_mean_sweep
            )
            self.context.recorded_thresholds_noise_std.append(
                processor_result.recorded_threshold_noise_std
            )
        self.context.recorded_threshold_session_config_used = self.session_config

    def calibrate_noise(self) -> None:
        """Estimates the standard deviation of the noise in each subsweep by setting enable_tx to
        False and collecting data, used to calculate the deviation.

        The calibration procedure can be done at any time as it is performed with Tx off, and is
        not effected by objects in front of the sensor.

        This function is called from the start() in the case of CFAR and Fixed threshold and from
        record_threshold() in the case of Recorded threshold. The reason for calling from
        record_threshold() is that it is used when calculating the threshold.
        """
        self._validate_ready_for_calibration()

        session_config = copy.deepcopy(self.session_config)

        for group in session_config.groups:
            group[self.sensor_id].sweeps_per_frame = 1
            for subsweep in group[self.sensor_id].subsweeps:
                subsweep.enable_tx = False
                subsweep.step_length = 1
                subsweep.start_point = 0
                # Set num_points to a high number to get sufficient number of data points to
                # estimate the standard deviation.
                subsweep.num_points = 500

        extended_metadata = self.client.setup_session(session_config)
        assert isinstance(extended_metadata, list)

        self.client.start_session()
        extended_result = self.client.get_next()
        assert isinstance(extended_result, list)
        self.client.stop_session()

        bg_noise_std = []
        for spec in self.processor_specs:
            bg_noise_std_in_subsweep = []
            result = extended_result[spec.group_index][spec.sensor_id]
            sensor_config = session_config.groups[spec.group_index][spec.sensor_id]
            subsweep_configs = sensor_config.subsweeps
            for idx in spec.subsweep_indexes:
                if not subsweep_configs[idx].enable_loopback:
                    subframe = result.subframes[idx]
                    subsweep_std = calculate_bg_noise_std(subframe, subsweep_configs[idx])
                    bg_noise_std_in_subsweep.append(subsweep_std)
            bg_noise_std.append(bg_noise_std_in_subsweep)
        self.context.bg_noise_std = bg_noise_std

    def calibrate_offset(self) -> None:
        self._validate_ready_for_calibration()

        sensor_config = a121.SensorConfig(
            start_point=-30,
            num_points=50,
            step_length=1,
            profile=a121.Profile.PROFILE_1,
            hwaas=64,
            sweeps_per_frame=1,
            enable_loopback=True,
            phase_enhancement=True,
        )

        session_config = a121.SessionConfig({self.sensor_id: sensor_config})
        self.client.setup_session(session_config)
        self.client.start_session()
        result = self.client.get_next()
        self.client.stop_session()

        assert isinstance(result, a121.Result)

        self.context.offset_m = calculate_offset(result, sensor_config)

    @classmethod
    def get_detector_status(
        cls, config: DetectorConfig, context: DetectorContext
    ) -> DetectorStatus:
        if not cls._valid_detector_config_range(config=config):
            return DetectorStatus(
                detector_state=DetailedStatus.INVALID_DETECTOR_CONFIG_RANGE,
                ready_to_calibrate_close_range=False,
                ready_to_record_threshold=False,
                ready_to_start=False,
            )

        (
            session_config,
            _,
        ) = cls._detector_to_session_config_and_processor_specs(config=config, sensor_id=1)

        ready_to_record_threshold = False
        if cls._has_close_range_measurement(config):
            ready_to_calibrate_close_range = True
            if cls._close_range_calibrated(context):
                if session_config != context.close_range_session_config_used:
                    detector_state = DetailedStatus.CLOSE_RANGE_CALIBRATION_CONFIG_MISMATCH
                elif not cls._recorded_threshold_calibrated(context):
                    detector_state = DetailedStatus.RECORDED_THRESHOLD_MISSING
                    ready_to_record_threshold = True
                elif session_config != context.recorded_threshold_session_config_used:
                    detector_state = DetailedStatus.RECORDED_THRESHOLD_CONFIG_MISMATCH
                else:
                    detector_state = DetailedStatus.OK
                    ready_to_record_threshold = True
            else:
                detector_state = DetailedStatus.CLOSE_RANGE_CALIBRATION_MISSING
        else:
            ready_to_calibrate_close_range = False
            if cls._has_recorded_threshold_mode(config):
                ready_to_record_threshold = True
                if cls._recorded_threshold_calibrated(context):
                    if session_config != context.recorded_threshold_session_config_used:
                        detector_state = DetailedStatus.RECORDED_THRESHOLD_CONFIG_MISMATCH
                    else:
                        detector_state = DetailedStatus.OK
                else:
                    detector_state = DetailedStatus.RECORDED_THRESHOLD_MISSING
            else:
                detector_state = DetailedStatus.OK

        return DetectorStatus(
            detector_state=detector_state,
            ready_to_calibrate_close_range=ready_to_calibrate_close_range,
            ready_to_record_threshold=ready_to_record_threshold,
            ready_to_start=(detector_state == DetailedStatus.OK),
        )

    @classmethod
    def _valid_detector_config_range(cls, config: DetectorConfig) -> bool:
        return cls.MIN_DIST_M < config.start_m and config.end_m < cls.MAX_DIST_M

    @staticmethod
    def _close_range_calibrated(context: DetectorContext) -> bool:
        has_dl = context.direct_leakage is not None
        has_pjcr = context.phase_jitter_comp_reference is not None

        if has_dl != has_pjcr:
            raise RuntimeError

        return has_dl and has_pjcr

    @staticmethod
    def _recorded_threshold_calibrated(context: DetectorContext) -> bool:
        return (
            context.recorded_thresholds_mean_sweep is not None
            and context.recorded_thresholds_noise_std is not None
        )

    @classmethod
    def _has_close_range_measurement(self, config: DetectorConfig) -> bool:
        (
            _,
            specs,
        ) = self._detector_to_session_config_and_processor_specs(config=config, sensor_id=1)
        return MeasurementType.CLOSE_RANGE in [
            spec.processor_config.measurement_type for spec in specs
        ]

    @classmethod
    def _has_recorded_threshold_mode(self, config: DetectorConfig) -> bool:
        (
            _,
            processor_specs,
        ) = self._detector_to_session_config_and_processor_specs(config=config, sensor_id=1)
        return ThresholdMethod.RECORDED in [
            spec.processor_config.threshold_method for spec in processor_specs
        ]

    def start(
        self, recorder: Optional[a121.Recorder] = None, skip_calibration: bool = False
    ) -> None:
        if self.started:
            raise RuntimeError("Already started")

        status = self.get_detector_status(self.detector_config, self.context)

        if not status.ready_to_start:
            raise RuntimeError("Not ready to start")

        if not skip_calibration:
            self.calibrate_noise()
            self.calibrate_offset()

        specs = self._add_context_to_processor_spec(self.processor_specs)
        extended_metadata = self.client.setup_session(self.session_config)
        assert isinstance(extended_metadata, list)
        assert self.context.offset_m is not None
        aggregator_config = AggregatorConfig(
            peak_sorting_method=self.detector_config.peaksorting_method
        )
        aggregator_context = AggregatorContext(offset_m=self.context.offset_m)
        self.aggregator = Aggregator(
            session_config=self.session_config,
            extended_metadata=extended_metadata,
            config=aggregator_config,
            context=aggregator_context,
            specs=specs,
        )

        if recorder is not None:
            if isinstance(recorder, a121.H5Recorder):
                algo_group = recorder.require_algo_group("distance_detector")
                _record_algo_data(
                    algo_group,
                    self.sensor_id,
                    self.detector_config,
                    self.context,
                )
            else:
                # Should never happen as we currently only have the H5Recorder
                warnings.warn("Will not save algo data")

        self.client.start_session(recorder)
        self.started = True

    def get_next(self) -> DetectorResult:
        if not self.started:
            raise RuntimeError("Not started")

        assert self.aggregator is not None

        extended_result = self.client.get_next()
        assert isinstance(extended_result, list)

        aggregator_result = self.aggregator.process(extended_result=extended_result)

        return DetectorResult(
            distances=aggregator_result.estimated_distances,
            processor_results=aggregator_result.processor_results,
            service_extended_result=aggregator_result.service_extended_result,
        )

    def update_config(self, config: DetectorConfig) -> None:
        (
            self.session_config,
            self.processor_specs,
        ) = self._detector_to_session_config_and_processor_specs(
            config=config, sensor_id=self.sensor_id
        )

    def stop(self) -> Any:
        if not self.started:
            raise RuntimeError("Already stopped")

        recorder_result = self.client.stop_session()

        self.started = False

        return recorder_result

    @classmethod
    def _detector_to_session_config_and_processor_specs(
        cls, config: DetectorConfig, sensor_id: int
    ) -> Tuple[a121.SessionConfig, list[ProcessorSpec]]:
        processor_specs = []
        groups = []
        group_index = 0

        plans = cls._create_group_plans(config)

        if MeasurementType.CLOSE_RANGE in plans:
            sensor_config = cls._close_subsweep_group_plans_to_sensor_config(
                plans[MeasurementType.CLOSE_RANGE]
            )
            groups.append({sensor_id: sensor_config})
            processor_specs.append(
                ProcessorSpec(
                    processor_config=ProcessorConfig(
                        threshold_method=ThresholdMethod.RECORDED,
                        measurement_type=MeasurementType.CLOSE_RANGE,
                        threshold_sensitivity=config.threshold_sensitivity,
                    ),
                    group_index=group_index,
                    sensor_id=sensor_id,
                    subsweep_indexes=[0, 1],
                )
            )
            group_index += 1

        if MeasurementType.FAR_RANGE in plans:
            (
                sensor_config,
                processor_specs_subsweep_indexes,
            ) = cls._far_subsweep_group_plans_to_sensor_config_and_subsweep_indexes(
                plans[MeasurementType.FAR_RANGE]
            )
            groups.append({sensor_id: sensor_config})

            processor_config = ProcessorConfig(
                threshold_method=config.threshold_method,
                fixed_threshold_value=config.fixed_threshold_value,
                threshold_sensitivity=config.threshold_sensitivity,
                cfar_one_sided=config.cfar_one_sided,
            )

            for subsweep_indexes in processor_specs_subsweep_indexes:
                processor_specs.append(
                    ProcessorSpec(
                        processor_config=processor_config,
                        group_index=group_index,
                        sensor_id=sensor_id,
                        subsweep_indexes=subsweep_indexes,
                    )
                )

        return (a121.SessionConfig(groups, extended=True), processor_specs)

    @classmethod
    def _create_group_plans(
        cls, config: DetectorConfig
    ) -> Dict[MeasurementType, List[SubsweepGroupPlan]]:
        """
        Create dictionary containing group plans for close and far range measurements.

        Constants used:
        - MIN_DIST_M defines the shortest distance possible to measure free of leakage.

        Outline of logic:
        - If the start is closer than the transition point, a close range subsweep plan is added.
        - If the end is further away than the transition point, far range subsweep plans are added.
        - max_profile is used in the far range region to achive high SNR and low power consumption.
        - A shorter profile is used between the transition point and the MIN_DIST_M of max_profile.
        """

        min_dist_m = cls._add_to_min_dist_m(config)
        transition_m = list(min_dist_m.values())[0]

        plans = {}

        if config.start_m < transition_m:
            profile = a121.Profile.PROFILE_1
            step_length = cls._limit_step_length(profile, config.max_step_length)
            breakpoints = cls._m_to_points([config.start_m, transition_m], step_length)
            hwaas = cls._calculate_hwaas(profile, breakpoints, config.signal_quality, step_length)

            has_neighbour = (False, transition_m < config.end_m)
            extended_breakpoints = cls._add_margin_to_breakpoints(
                profile=profile,
                step_length=step_length,
                base_bpts=breakpoints,
                has_neighbour=has_neighbour,
                config=config,
                is_close_range_measurement=True,
            )

            plans[MeasurementType.CLOSE_RANGE] = [
                SubsweepGroupPlan(
                    step_length=step_length,
                    breakpoints=extended_breakpoints,
                    profile=profile,
                    hwaas=hwaas,
                )
            ]

        far_subgroup_plans = []
        far_range_start_m = np.max([config.start_m, transition_m])
        if (
            config.max_profile is not a121.Profile.PROFILE_1
            and config.start_m < min_dist_m[config.max_profile]
            and far_range_start_m < config.end_m
        ):
            min_dists_m = np.array(list(min_dist_m.values()))
            min_dists_profiles = np.array(list(min_dist_m.keys()))
            (viable_profile_idx,) = np.where(min_dists_m <= far_range_start_m)
            profile_to_be_used = min_dists_profiles[viable_profile_idx[-1]]

            end_m = min(config.end_m, min_dist_m[config.max_profile])
            step_length = cls._limit_step_length(profile_to_be_used, config.max_step_length)
            breakpoints = cls._m_to_points([far_range_start_m, end_m], step_length)
            hwaas = cls._calculate_hwaas(
                profile_to_be_used, breakpoints, config.signal_quality, step_length
            )

            has_neighbour = (len(plans) != 0, min_dist_m[config.max_profile] < end_m)
            extended_breakpoints = cls._add_margin_to_breakpoints(
                profile=profile_to_be_used,
                step_length=step_length,
                base_bpts=breakpoints,
                has_neighbour=has_neighbour,
                config=config,
                is_close_range_measurement=False,
            )

            far_subgroup_plans.append(
                SubsweepGroupPlan(
                    step_length=step_length,
                    breakpoints=extended_breakpoints,
                    profile=profile_to_be_used,
                    hwaas=hwaas,
                )
            )

        if min_dist_m[config.max_profile] < config.end_m:
            subsweep_start_m = max([config.start_m, min_dist_m[config.max_profile]])
            breakpoints_m = np.linspace(
                subsweep_start_m,
                config.end_m,
                cls.NUM_SUBSWEEPS_IN_SENSOR_CONFIG + 1 - len(far_subgroup_plans),
            ).tolist()

            profile = config.max_profile
            step_length = cls._limit_step_length(config.max_profile, config.max_step_length)
            breakpoints = cls._m_to_points(breakpoints_m, step_length)
            hwaas = cls._calculate_hwaas(profile, breakpoints, config.signal_quality, step_length)

            has_neighbour = (len(plans) != 0 or len(far_subgroup_plans) != 0, False)
            extended_breakpoints = cls._add_margin_to_breakpoints(
                profile=profile,
                step_length=step_length,
                base_bpts=breakpoints,
                has_neighbour=has_neighbour,
                config=config,
                is_close_range_measurement=False,
            )

            far_subgroup_plans.append(
                SubsweepGroupPlan(
                    step_length=step_length,
                    breakpoints=extended_breakpoints,
                    profile=profile,
                    hwaas=hwaas,
                )
            )

        if len(far_subgroup_plans) != 0:
            plans[MeasurementType.FAR_RANGE] = far_subgroup_plans

        return plans

    @classmethod
    def _add_to_min_dist_m(cls, config: DetectorConfig) -> Dict[a121.Profile, float]:
        min_dist_m = {}
        for profile, min_dist in cls.MIN_LEAKAGE_FREE_DIST_M.items():
            min_dist_m[profile] = min_dist
            if config.threshold_method == ThresholdMethod.CFAR:
                step_length = cls._limit_step_length(profile, config.max_step_length)
                cfar_margin_m = (
                    Processor.calc_cfar_margin(profile, step_length)
                    * step_length
                    * Processor.APPROX_BASE_STEP_LENGTH_M
                )
                min_dist_m[profile] += cfar_margin_m
        return min_dist_m

    @classmethod
    def _calculate_hwaas(
        cls, profile: a121.Profile, breakpoints: list[int], signal_quality: float, step_length: int
    ) -> list[int]:
        rlg_per_hwaas = cls.RLG_PER_HWAAS_MAP[profile]
        hwaas = []
        for idx in range(len(breakpoints) - 1):
            processing_gain = cls._calc_processing_gain(profile, step_length)
            subsweep_end_point_m = max(
                Processor.APPROX_BASE_STEP_LENGTH_M * breakpoints[idx + 1],
                cls.HWAAS_MIN_DISTANCE,
            )
            rlg = signal_quality + 40 * np.log10(subsweep_end_point_m) - np.log10(processing_gain)
            hwaas_in_subsweep = int(10 ** ((rlg - rlg_per_hwaas) / 10))
            hwaas.append(np.clip(hwaas_in_subsweep, cls.MIN_HWAAS, cls.MAX_HWAAS))
        return hwaas

    @staticmethod
    def _calc_processing_gain(profile: a121.Profile, step_length: int) -> float:
        envelope_base_length_m = Processor.ENVELOPE_FWHM_M[profile] * 2  # approx envelope width
        num_points_in_envelope = int(
            envelope_base_length_m / (step_length * Processor.APPROX_BASE_STEP_LENGTH_M)
        )
        mid_point = num_points_in_envelope // 2
        pulse = np.concatenate(
            (
                np.linspace(0, 1, mid_point),
                np.linspace(1, 0, num_points_in_envelope - mid_point),
            )
        )
        return float(np.sum(pulse**2))

    @classmethod
    def _add_margin_to_breakpoints(
        cls,
        profile: a121.Profile,
        step_length: int,
        base_bpts: list[int],
        has_neighbour: Tuple[bool, bool],
        config: DetectorConfig,
        is_close_range_measurement: bool,
    ) -> list[int]:
        """
        Add points to segment edges based on their position.

        1. Add one margin to each segment for distance filter initialization
        2. Add an additional margin to segments with neighbouring segments for segment overlap
        """

        margin_p = Processor.distance_filter_edge_margin(profile, step_length) * step_length
        left_margin = margin_p
        right_margin = margin_p

        if has_neighbour[0]:
            left_margin += margin_p

        if has_neighbour[1]:
            right_margin += margin_p

        if config.threshold_method == ThresholdMethod.CFAR and not is_close_range_measurement:
            cfar_margin = Processor.calc_cfar_margin(profile, step_length) * step_length
            left_margin += cfar_margin
            right_margin += cfar_margin

        bpts = copy.copy(base_bpts)
        bpts[0] -= left_margin
        bpts[-1] += right_margin

        return bpts

    @classmethod
    def _limit_step_length(cls, profile: a121.Profile, user_limit: Optional[int]) -> int:
        fwhm_p = Processor.ENVELOPE_FWHM_M[profile] / Processor.APPROX_BASE_STEP_LENGTH_M
        limit = int(fwhm_p / cls.MIN_NUM_POINTS_IN_ENVELOPE_FWHM_SPAN)

        if user_limit is not None:
            limit = min(user_limit, limit)

        if limit < cls.NUM_POINTS_IN_COARSE:
            # TODO: Pick the longest, but not longer than limit
            idx_closest = np.argmin(np.abs(np.array(cls.VALID_STEP_LENGTHS_IN_COARSE) - limit))
            return cls.VALID_STEP_LENGTHS_IN_COARSE[idx_closest]
        else:
            return (limit // cls.NUM_POINTS_IN_COARSE) * cls.NUM_POINTS_IN_COARSE

    @classmethod
    def _close_subsweep_group_plans_to_sensor_config(
        cls, plan_: List[SubsweepGroupPlan]
    ) -> a121.SensorConfig:
        (plan,) = plan_
        subsweeps = []
        subsweeps.append(
            a121.SubsweepConfig(
                start_point=0,
                num_points=1,
                step_length=1,
                profile=a121.Profile.PROFILE_4,
                hwaas=plan.hwaas[0],
                receiver_gain=15,
                phase_enhancement=True,
                enable_loopback=True,
            )
        )
        num_points = int((plan.breakpoints[1] - plan.breakpoints[0]) / plan.step_length)
        subsweeps.append(
            a121.SubsweepConfig(
                start_point=plan.breakpoints[0],
                num_points=num_points,
                step_length=plan.step_length,
                profile=plan.profile,
                hwaas=plan.hwaas[0],
                receiver_gain=5,
                phase_enhancement=True,
                prf=cls._select_prf(plan.breakpoints[1], plan.profile),
            )
        )
        return a121.SensorConfig(subsweeps=subsweeps, sweeps_per_frame=10)

    @classmethod
    def _far_subsweep_group_plans_to_sensor_config_and_subsweep_indexes(
        cls, subsweep_group_plans: list[SubsweepGroupPlan]
    ) -> Tuple[a121.SensorConfig, list[list[int]]]:
        subsweeps = []
        processor_specs_subsweep_indexes = []
        subsweep_idx = 0
        for plan in subsweep_group_plans:
            subsweep_indexes = []
            for bp_idx in range(len(plan.breakpoints) - 1):
                num_points = int(
                    (plan.breakpoints[bp_idx + 1] - plan.breakpoints[bp_idx]) / plan.step_length
                )
                subsweeps.append(
                    a121.SubsweepConfig(
                        start_point=plan.breakpoints[bp_idx],
                        num_points=num_points,
                        step_length=plan.step_length,
                        profile=plan.profile,
                        hwaas=plan.hwaas[bp_idx],
                        receiver_gain=10,
                        phase_enhancement=True,
                        prf=cls._select_prf(plan.breakpoints[bp_idx + 1], plan.profile),
                    )
                )
                subsweep_indexes.append(subsweep_idx)
                subsweep_idx += 1
            processor_specs_subsweep_indexes.append(subsweep_indexes)
        return (
            a121.SensorConfig(subsweeps=subsweeps, sweeps_per_frame=1),
            processor_specs_subsweep_indexes,
        )

    @classmethod
    def _select_prf(cls, breakpoint: int, profile: a121.Profile) -> a121.PRF:
        max_meas_dist_m = copy.copy(cls.MAX_MEASURABLE_DIST_M)

        if (
            a121.PRF.PRF_19_5_MHz in max_meas_dist_m
            and profile != a121.Profile.PROFILE_1
            and profile != a121.Profile.PROFILE_2
        ):
            del max_meas_dist_m[a121.PRF.PRF_19_5_MHz]

        breakpoint_m = breakpoint * Processor.APPROX_BASE_STEP_LENGTH_M
        viable_prfs = [
            prf for prf, max_dist_m in max_meas_dist_m.items() if breakpoint_m < max_dist_m
        ]
        return sorted(viable_prfs, key=lambda prf: prf.frequency)[-1]

    @classmethod
    def _m_to_points(cls, breakpoints_m: list[float], step_length: int) -> list[int]:
        bpts_m = np.array(breakpoints_m)
        start_point = int(bpts_m[0] / Processor.APPROX_BASE_STEP_LENGTH_M)
        num_steps = (bpts_m[-1] - bpts_m[0]) / (Processor.APPROX_BASE_STEP_LENGTH_M)
        bpts = num_steps / (bpts_m[-1] - bpts_m[0]) * (bpts_m - bpts_m[0]) + start_point
        return [(bpt // step_length) * step_length for bpt in bpts]

    @classmethod
    def _update_processor_mode(
        cls, processor_specs: list[ProcessorSpec], processor_mode: ProcessorMode
    ) -> list[ProcessorSpec]:
        updated_specs = []
        for spec in processor_specs:
            new_processor_config = attrs.evolve(
                spec.processor_config, processor_mode=processor_mode
            )
            updated_specs.append(attrs.evolve(spec, processor_config=new_processor_config))
        return updated_specs

    @classmethod
    def _filter_close_range_spec(cls, specs: list[ProcessorSpec]) -> list[ProcessorSpec]:
        NUM_CLOSE_RANGE_SPECS = 1
        close_range_specs = []
        for spec in specs:
            if spec.processor_config.measurement_type == MeasurementType.CLOSE_RANGE:
                close_range_specs.append(spec)
        if len(close_range_specs) != NUM_CLOSE_RANGE_SPECS:
            raise ValueError("Incorrect subsweep config for close range measurement")

        return close_range_specs

    def _add_context_to_processor_spec(
        self, processor_specs: list[ProcessorSpec]
    ) -> list[ProcessorSpec]:
        """
        Add context to ProcessorSpec object, based on the corresponding processor config.

        1. Close range measurement and distance estimation -> Recorded threshold, direct
        leakage and phase jitter compensation reference.
        2. Close range measurement and recorded threshold calibration -> Direct leakage and
        phase jitter compensation reference.
        3. Far range measurement and recorded threshold -> Recorded threshold.

        If not one of these cases, add the unaltered processor specification.
        """

        ERR_MESSAGE_CLOSE_RANGE_ERR = "Close range calibration not performed"
        ERR_MESSAGE_RECORDED = "Recorded threshold calibration not performed"

        assert self.context.bg_noise_std is not None

        updated_specs: List[ProcessorSpec] = []

        for idx, (spec, bg_noise_std) in enumerate(
            zip(processor_specs, self.context.bg_noise_std)
        ):

            if (
                spec.processor_config.measurement_type == MeasurementType.CLOSE_RANGE
                and spec.processor_config.processor_mode == ProcessorMode.DISTANCE_ESTIMATION
            ):
                if not self._close_range_calibrated(self.context):
                    raise Exception(ERR_MESSAGE_CLOSE_RANGE_ERR)

                if (
                    not self._recorded_threshold_calibrated(self.context)
                    or self.context.recorded_thresholds_mean_sweep is None
                    or self.context.recorded_thresholds_noise_std is None
                ):
                    raise Exception(ERR_MESSAGE_RECORDED)

                context = ProcessorContext(
                    recorded_threshold_mean_sweep=self.context.recorded_thresholds_mean_sweep[idx],
                    recorded_threshold_noise_std=self.context.recorded_thresholds_noise_std[idx],
                    bg_noise_std=bg_noise_std,
                    direct_leakage=self.context.direct_leakage,
                    phase_jitter_comp_ref=self.context.phase_jitter_comp_reference,
                )
                updated_specs.append(attrs.evolve(spec, processor_context=context))
            elif (
                spec.processor_config.measurement_type == MeasurementType.CLOSE_RANGE
                and spec.processor_config.processor_mode
                == ProcessorMode.RECORDED_THRESHOLD_CALIBRATION
            ):
                if not self._close_range_calibrated(self.context):
                    raise Exception(ERR_MESSAGE_CLOSE_RANGE_ERR)

                context = ProcessorContext(
                    bg_noise_std=bg_noise_std,
                    direct_leakage=self.context.direct_leakage,
                    phase_jitter_comp_ref=self.context.phase_jitter_comp_reference,
                )
                updated_specs.append(attrs.evolve(spec, processor_context=context))
            elif (
                spec.processor_config.measurement_type == MeasurementType.FAR_RANGE
                and spec.processor_config.threshold_method == ThresholdMethod.RECORDED
                and self._recorded_threshold_calibrated(self.context)
                and self.context.recorded_threshold_session_config_used == self.session_config
            ):
                if (
                    self.context.recorded_thresholds_mean_sweep is None
                    or self.context.recorded_thresholds_noise_std is None
                ):
                    raise Exception(ERR_MESSAGE_RECORDED)

                context = ProcessorContext(
                    recorded_threshold_mean_sweep=self.context.recorded_thresholds_mean_sweep[idx],
                    recorded_threshold_noise_std=self.context.recorded_thresholds_noise_std[idx],
                    bg_noise_std=bg_noise_std,
                )
                updated_specs.append(attrs.evolve(spec, processor_context=context))
            elif (
                spec.processor_config.measurement_type == MeasurementType.FAR_RANGE
                and spec.processor_config.processor_mode
                == ProcessorMode.RECORDED_THRESHOLD_CALIBRATION
            ):
                context = ProcessorContext(bg_noise_std=bg_noise_std)
                updated_specs.append(attrs.evolve(spec, processor_context=context))
            else:
                context = ProcessorContext(bg_noise_std=bg_noise_std)
                updated_specs.append(attrs.evolve(spec, processor_context=context))
        return updated_specs


def _record_algo_data(
    algo_group: h5py.Group,
    sensor_id: int,
    detector_config: DetectorConfig,
    context: DetectorContext,
) -> None:
    algo_group.create_dataset(
        "sensor_id",
        data=sensor_id,
        track_times=False,
    )
    algo_group.create_dataset(
        "detector_config",
        data=detector_config.to_json(),
        dtype=a121._H5PY_STR_DTYPE,
        track_times=False,
    )

    context_group = algo_group.create_group("context")

    for k, v in attrs.asdict(context).items():
        if k in [
            "recorded_thresholds_mean_sweep",
            "recorded_thresholds_noise_std",
            "bg_noise_std",
        ]:
            continue

        if v is None:
            continue

        if isinstance(v, a121.SessionConfig):
            context_group.create_dataset(
                k,
                data=v.to_json(),
                dtype=a121._H5PY_STR_DTYPE,
                track_times=False,
            )
        elif isinstance(v, np.ndarray) or isinstance(v, float) or isinstance(v, int):
            context_group.create_dataset(k, data=v, track_times=False)
        else:
            raise RuntimeError(f"Unexpected {DetectorContext.__name__} field type '{type(v)}'")

    if context.recorded_thresholds_mean_sweep is not None:
        recorded_thresholds_mean_sweep_group = context_group.create_group(
            "recorded_thresholds_mean_sweep"
        )

        for i, v in enumerate(context.recorded_thresholds_mean_sweep):
            recorded_thresholds_mean_sweep_group.create_dataset(
                f"index_{i}", data=v, track_times=False
            )

    if context.recorded_thresholds_noise_std is not None:
        recorded_thresholds_std_group = context_group.create_group("recorded_thresholds_noise_std")

        for i, v in enumerate(context.recorded_thresholds_noise_std):
            recorded_thresholds_std_group.create_dataset(f"index_{i}", data=v, track_times=False)

    if context.bg_noise_std is not None:
        bg_noise_std_group = context_group.create_group("bg_noise_std")

        for i, v in enumerate(context.bg_noise_std):
            bg_noise_std_group.create_dataset(f"index_{i}", data=v, track_times=False)


def _load_algo_data(algo_group: h5py.Group) -> Tuple[int, DetectorConfig, DetectorContext]:
    sensor_id = algo_group["sensor_id"][()]
    config = DetectorConfig.from_json(algo_group["detector_config"][()])

    context_dict = {}
    context_group = algo_group["context"]

    unknown_keys = set(context_group.keys()) - set(attrs.fields_dict(DetectorContext).keys())
    if unknown_keys:
        raise Exception(f"Unknown field(s) in stored context: {unknown_keys}")

    field_map = {
        "offset_m": None,
        "direct_leakage": None,
        "phase_jitter_comp_reference": None,
        "recorded_threshold_session_config_used": a121.SessionConfig.from_json,
        "close_range_session_config_used": a121.SessionConfig.from_json,
    }
    for k, func in field_map.items():
        try:
            v = context_group[k][()]
        except KeyError:
            continue

        context_dict[k] = func(v) if func else v

    if "recorded_thresholds_mean_sweep" in context_group:
        mean_sweeps = _get_group_items(context_group["recorded_thresholds_mean_sweep"])
        context_dict["recorded_thresholds_mean_sweep"] = mean_sweeps

    if "recorded_thresholds_noise_std" in context_group:
        noise_stds = _get_group_items(context_group["recorded_thresholds_noise_std"])
        context_dict["recorded_thresholds_noise_std"] = noise_stds

    if "bg_noise_std" in context_group:
        bg_noise_std = _get_group_items(context_group["bg_noise_std"])
        context_dict["bg_noise_std"] = bg_noise_std

    context = DetectorContext(**context_dict)

    return sensor_id, config, context


def _get_group_items(group: h5py.Group) -> list[npt.NDArray]:
    group_items = []

    i = 0
    while True:
        try:
            v = group[f"index_{i}"][()]
        except KeyError:
            break

        group_items.append(v)
        i += 1
    return group_items



# Copyright (c) Acconeer AB, 2022
# All rights reserved

SDK_VERSION = "0.4.1"

from ._cli import ExampleArgumentParser, get_client_args
from ._core import (
    _H5PY_STR_DTYPE,
    PRF,
    Client,
    ClientError,
    ClientInfo,
    H5Record,
    H5Recorder,
    IdleState,
    InMemoryRecord,
    Metadata,
    PersistentRecord,
    Profile,
    Record,
    Recorder,
    Result,
    SensorConfig,
    SensorInfo,
    ServerError,
    ServerInfo,
    SessionConfig,
    StackedResults,
    SubsweepConfig,
    ValidationError,
    ValidationResult,
    ValidationWarning,
    iterate_extended_structure,
    load_record,
    open_record,
    save_record,
    save_record_to_h5,
)
from ._core_ext import _ReplayingClient, _StopReplay
from ._perf_calc import _PerformanceCalc



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from __future__ import annotations

import json
import typing as t
import warnings

import attrs

from acconeer.exptool.a121._core.utils import (
    EntityJSONEncoder,
    convert_value,
    is_divisor_of,
    is_multiple_of,
    pretty_dict_line_strs,
)

from .config_enums import PRF, Profile
from .validation_error import ValidationError, ValidationResult, ValidationWarning


SPARSE_IQ_PPC = 24

# The `converter` argument to `attrs.field` influences the signature of `__init__`
# - (https://www.attrs.org/en/stable/api.html#converters)
# This is mirrored in LSPs and the docs.
#
# `convert_value` is a generic function which needs to be specialiced to `int`s for use here
# Plain `Profile` / `PRF` functions as converters in runtime, but no type is in
# the signature of `__init__`.


def int_converter(value: int) -> int:
    return convert_value(value, factory=int)


def profile_converter(profile: Profile) -> Profile:
    return Profile(profile)


def prf_converter(prf: PRF) -> PRF:
    return PRF(prf)


@attrs.mutable(kw_only=True)
class SubsweepConfig:
    """Subsweep configuration

    The subsweep config represents a 1-1 mapping to the RSS service subsweep config.

    Normally used as a part of the :attr:`SensorConfig`.
    """

    _start_point: int = attrs.field(
        default=80,
        converter=int_converter,
    )
    _num_points: int = attrs.field(
        default=160,
        converter=int_converter,
        validator=[attrs.validators.ge(1)],
    )
    _step_length: int = attrs.field(
        default=1,
        converter=int_converter,
    )
    _profile: Profile = attrs.field(
        default=Profile.PROFILE_3,
        converter=profile_converter,
    )
    _hwaas: int = attrs.field(
        default=8,
        converter=int_converter,
        validator=[attrs.validators.ge(1), attrs.validators.le(511)],
    )
    _receiver_gain: int = attrs.field(
        default=16,
        converter=int_converter,
        validator=[attrs.validators.ge(0), attrs.validators.le(23)],
    )
    _enable_tx: bool = attrs.field(
        default=True,
        converter=bool,
    )
    _enable_loopback: bool = attrs.field(
        default=False,
        converter=bool,
    )
    _phase_enhancement: bool = attrs.field(
        default=False,
        converter=bool,
    )
    _prf: PRF = attrs.field(
        default=PRF.PRF_13_0_MHz,
        converter=prf_converter,
    )

    def _collect_validation_results(self) -> list[ValidationResult]:
        validation_results: list[ValidationResult] = []

        if self.enable_loopback and self.profile == Profile.PROFILE_2:
            validation_results.extend(
                [
                    ValidationError(
                        self, "enable_loopback", "Enable loopback is incompatible with Profile 2."
                    ),
                    ValidationError(
                        self, "profile", "Enable loopback is incompatible with Profile 2."
                    ),
                ]
            )

        if self.prf == PRF.PRF_19_5_MHz:
            FORBIDDEN_PROFILES = [Profile.PROFILE_3, Profile.PROFILE_4, Profile.PROFILE_5]

            if self.profile in FORBIDDEN_PROFILES:
                validation_results.extend(
                    [
                        ValidationError(
                            self, "prf", "19.5 MHz PRF is only compatible with profile 1 and 2."
                        ),
                        ValidationError(
                            self,
                            "profile",
                            f"Profile {self.profile.value} is not supported with 19.5 MHz PRF.",
                        ),
                    ]
                )

        return validation_results

    def validate(self) -> None:
        """Performs self-validation

        :raises ValidationError: If anything is invalid.
        """
        for validation_result in self._collect_validation_results():
            try:
                raise validation_result
            except ValidationWarning as vw:
                warnings.warn(vw.message)

    @property
    def start_point(self) -> int:
        """Starting point of the sweep

        The starting point of the sweep. The corresponding start in millimeter is approximately
        ``start_point`` * 2.5 mm.
        """

        return self._start_point

    @start_point.setter
    def start_point(self, value: int) -> None:
        self._start_point = value

    @property
    def num_points(self) -> int:
        """Number of data points to measure

        The number of data points to measure in a sweep.
        """

        return self._num_points

    @num_points.setter
    def num_points(self, value: int) -> None:
        self._num_points = value

    @property
    def step_length(self) -> int:
        """Step length in a sweep

        This sets the number of steps to have between each data point.

        Sampling produces complex (IQ) data points with configurable distance spacing, starting
        from ~2.5mm.
        """

        return self._step_length

    @step_length.setter
    def step_length(self, value: int) -> None:
        self._step_length = value

    @_step_length.validator
    def _(self, _: attrs.Attribute, step_length: int) -> None:
        if not (
            is_divisor_of(SPARSE_IQ_PPC, step_length) or is_multiple_of(SPARSE_IQ_PPC, step_length)
        ):
            raise ValueError(f"step_length must be a divisor or multiple of {SPARSE_IQ_PPC}")

    @property
    def profile(self) -> Profile:
        """Profile

        Each profile consists of a number of settings for the sensor that configures the RX and TX
        paths. Lower profiles have higher depth resolution while higher profiles have higher radar
        loop gain.

        See also :class:`Profile`.
        """

        return self._profile

    @profile.setter
    def profile(self, value: Profile) -> None:
        self._profile = value

    @property
    def hwaas(self) -> int:
        """Hardware accelerated average samples (HWAAS)

        Each data point can be sampled several times and the sensor hardware then produces an
        average value of those samples. The time needed to measure a sweep is roughly proportional
        to the number of averaged samples. Hence, if there is a need to obtain a higher update
        rate, HWAAS could be decreased but this leads to lower SNR.

        HWAAS must be between 1 and 511 inclusive.
        """

        return self._hwaas

    @hwaas.setter
    def hwaas(self, value: int) -> None:
        self._hwaas = value

    @property
    def receiver_gain(self) -> int:
        """Receiver gain setting

        Must be a value between 0 and 23 inclusive where 23 is the highest gain and 0 the lowest.

        Lower gain gives higher SNR. However, too low gain may result in quantization, lowering
        SNR. Too high gain may result in saturation, corrupting the data.
        """

        return self._receiver_gain

    @receiver_gain.setter
    def receiver_gain(self, value: int) -> None:
        self._receiver_gain = value

    @property
    def enable_tx(self) -> bool:
        """Enable or disable the transmitter

        If set to True, TX is enabled. This will enable the radio transmitter. By turning the
        transmitter off the RX noise floor can be measured.
        """

        return self._enable_tx

    @enable_tx.setter
    def enable_tx(self, value: bool) -> None:
        self._enable_tx = value

    @property
    def enable_loopback(self) -> bool:
        """Enable or disable loopback

        Note, loopback can't be enabled together with profile 2.
        """
        return self._enable_loopback

    @enable_loopback.setter
    def enable_loopback(self, value: bool) -> None:
        self._enable_loopback = value

    @property
    def phase_enhancement(self) -> bool:
        """Enable or disable phase enhancement

        If enabled, the data phase will be enhanced such that coherent distance filtering can be
        applied. Given a single reflection from an object, the phase will appear as "flat" around
        the amplitude peak.

        Enabling the phase enhancement increases the processing execution time.
        """

        return self._phase_enhancement

    @phase_enhancement.setter
    def phase_enhancement(self, value: bool) -> None:
        self._phase_enhancement = value

    @property
    def prf(self) -> PRF:
        """Pulse Repetition Frequency (PRF)

        See :class:`PRF` for details.
        """

        return self._prf

    @prf.setter
    def prf(self, value: PRF) -> None:
        self._prf = PRF(value)

    def to_dict(self) -> dict[str, t.Any]:
        return {k.strip("_"): v for k, v in attrs.asdict(self).items()}

    @classmethod
    def from_dict(cls, d: dict) -> SubsweepConfig:
        return SubsweepConfig(**d)

    def to_json(self) -> str:
        return json.dumps(self.to_dict(), cls=EntityJSONEncoder)

    @classmethod
    def from_json(cls, json_str: str) -> SubsweepConfig:
        return cls.from_dict(json.loads(json_str))

    def _pretty_str_lines(self, index: t.Optional[int] = None) -> list[str]:
        lines = []
        index_str = "" if index is None else f" @ index {index}"
        lines.append(f"{type(self).__name__}{index_str}:")
        lines.extend(pretty_dict_line_strs(self.to_dict()))
        return lines

    def __str__(self) -> str:
        return "\n".join(self._pretty_str_lines())



# Copyright (c) Acconeer AB, 2022
# All rights reserved

import matplotlib.pyplot as plt
import numpy as np

import acconeer.exptool as et


def main():
    args = et.a111.ExampleArgumentParser(num_sens=1).parse_args()
    et.utils.config_logging(args)

    client = et.a111.Client(**et.a111.get_client_args(args))

    config = et.a111.IQServiceConfig()
    config.sensor = args.sensors
    config.update_rate = 10

    session_info = client.setup_session(config)
    depths = et.a111.get_range_depths(config, session_info)

    amplitude_y_max = 1000

    fig, (amplitude_ax, phase_ax) = plt.subplots(2)
    fig.set_size_inches(8, 6)
    fig.canvas.set_window_title("Acconeer matplotlib example")

    for ax in [amplitude_ax, phase_ax]:
        ax.set_xlabel("Depth (m)")
        ax.set_xlim(config.range_interval)
        ax.grid(True)

    amplitude_ax.set_ylabel("Amplitude")
    amplitude_ax.set_ylim(0, 1.1 * amplitude_y_max)
    phase_ax.set_ylabel("Phase")
    et.utils.mpl_setup_yaxis_for_phase(phase_ax)

    amplitude_line = amplitude_ax.plot(depths, np.zeros_like(depths))[0]
    phase_line = phase_ax.plot(depths, np.zeros_like(depths))[0]

    fig.tight_layout()
    plt.ion()
    plt.show()

    interrupt_handler = et.utils.ExampleInterruptHandler()
    print("Press Ctrl-C to end session")

    client.start_session()

    while not interrupt_handler.got_signal:
        info, data = client.get_next()

        amplitude = np.abs(data)
        phase = np.angle(data)

        max_amplitude = np.max(amplitude)
        if max_amplitude > amplitude_y_max:
            amplitude_y_max = max_amplitude
            amplitude_ax.set_ylim(0, 1.1 * max_amplitude)

        amplitude_line.set_ydata(amplitude)
        phase_line.set_ydata(phase)

        if not plt.fignum_exists(1):  # Simple way to check if plot is closed
            break

        fig.canvas.flush_events()

    print("Disconnecting...")
    plt.close()
    client.disconnect()


if __name__ == "__main__":
    main()



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path
from typing import Optional

import numpy as np

import acconeer.exptool as et
from acconeer.exptool.a111.algo import (
    Calibration,
    CalibrationMapper,
    _AcceptedFileExtensions,
    _Path,
)


@dataclass(frozen=True)
class DistanceDetectorCalibration(Calibration):
    stationary_clutter_mean: np.ndarray
    stationary_clutter_std: np.ndarray

    def save(self, path: _Path):
        self.validate_path(path, file_extensions=[self.file_extensions()[0]])

        np.savez(
            path,
            mean=self.stationary_clutter_mean,
            std=self.stationary_clutter_std,
        )

    @classmethod
    def load(cls, path: _Path) -> Calibration:
        cls.validate_path(path)

        # Distance Detector saved background as ".npy" in v3
        if Path(path).suffix == ".npy":
            mean_std = np.load(path)
            return DistanceDetectorCalibration(mean_std[0], mean_std[1])
        else:
            with np.load(path) as archive:
                return DistanceDetectorCalibration(
                    archive["mean"],
                    archive["std"],
                )

    @classmethod
    def file_extensions(cls) -> _AcceptedFileExtensions:
        return [
            ("npz", "Numpy data archives (*.npz)"),
            ("npy", "Numpy data files (*.npy)"),
        ]


class DistaceDetectorCalibrationConfiguration(et.configbase.Config):
    pass


class DistaceDetectorCalibrationMapper(CalibrationMapper):
    @classmethod
    def get_updated_calibration_from_configuration(
        cls,
        configuration: DistaceDetectorCalibrationConfiguration,
        calibration: Optional[DistanceDetectorCalibration],
    ) -> DistanceDetectorCalibration:
        if calibration is None:
            raise ValueError("Calibration cannot be None in this context.")

        return calibration

    @classmethod
    def update_config_from_calibration(
        cls,
        configuration: DistaceDetectorCalibrationConfiguration,
        calibration: DistanceDetectorCalibration,
    ) -> None:
        pass



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from acconeer.exptool.a111.algo import ModuleFamily, ModuleInfo
from acconeer.exptool.a111.algo.utils import multi_sensor_pg_updater, multi_sensor_processor

from ._processor import ProcessingConfiguration, Processor, get_sensor_config
from .ui import PGUpdater


module_info = ModuleInfo(
    key="sparse_presence",
    label="Presence detection (sparse)",
    pg_updater=multi_sensor_pg_updater(PGUpdater),
    processing_config_class=ProcessingConfiguration,
    module_family=ModuleFamily.DETECTOR,
    sensor_config_class=get_sensor_config,
    processor=multi_sensor_processor(Processor),
    multi_sensor=True,
    docs_url="https://docs.acconeer.com/en/latest/processing/presence_detection_sparse.html",
)



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from __future__ import annotations

from acconeer.exptool.a121._core.entities import PersistentRecord, Record
from acconeer.exptool.a121._core.peripherals.im_record import InMemoryRecord

from .record import H5Record
from .recorder import H5Recorder
from .utils import PathOrH5File, h5_file_factory


def open_record(path_or_file: PathOrH5File) -> PersistentRecord:
    """Open a record from file

    Since this function returns a :class:`PersistentRecord`, data is not immediately loaded into
    memory. Rather, data is lazily loaded from the underlying file on demand.

    Either a path or an opened file (:class:`h5py.File`) may be given. If a path is given, the file
    will be opened by this function, and must be closed again by the user. The recommended way to
    do this is by using the context manager of the returned persistent record, like:

    .. code-block:: python

        with a121.open_record("path/to/my/file.h5") as record:
            record.timestamp

    .. tip::

        Unless you're dealing with very large files, use :func:`load_record` instead.

    :returns: A :class:`PersistentRecord` wrapping the given file
    """

    file, _ = h5_file_factory(path_or_file, h5_file_mode="r")
    return H5Record(file)


def load_record(path_or_file: PathOrH5File) -> Record:
    """Load a record from file

    Unlike :func:`open_record`, this functions loads the data into memory immediately. The file
    handle is not kept open.

    :returns: A :class:`Record` with the content of the given file
    """

    with open_record(path_or_file) as h5_record:
        return InMemoryRecord.from_record(h5_record)


def save_record(path_or_file: PathOrH5File, record: Record) -> None:
    """Alias for :func:`save_record_to_h5`"""

    return save_record_to_h5(path_or_file, record)


def save_record_to_h5(path_or_file: PathOrH5File, record: Record) -> None:
    """Save a record to a HDF5 file"""

    recorder = H5Recorder(path_or_file)

    recorder._start(
        client_info=record.client_info,
        extended_metadata=record.extended_metadata,
        server_info=record.server_info,
        session_config=record.session_config,
    )

    for extended_result in record.extended_results:
        recorder._sample(extended_result)

    recorder._stop()



# Copyright (c) Acconeer AB, 2022
# All rights reserved

import argparse


class ExptoolArgumentParser(argparse.ArgumentParser):
    def __init__(self) -> None:
        super().__init__()

        self.add_argument("--amv", action="store_true")
        self.add_argument(
            "--portable",
            action="store_true",
            help=argparse.SUPPRESS,  # makes option hidden
        )
        self.add_argument(
            "--purge-config",
            action="store_true",
            help="Remove configuration files.",
        )

        verbosity_group = self.add_mutually_exclusive_group(required=False)
        verbosity_group.add_argument(
            "-v",
            "--verbose",
            action="store_true",
        )
        verbosity_group.add_argument(
            "-vv",
            "--debug",
            action="store_true",
        )
        verbosity_group.add_argument(
            "-q",
            "--quiet",
            action="store_true",
        )



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from __future__ import annotations

import numpy as np

# Added here to force pyqtgraph to choose PySide
import PySide6  # noqa: F401

import pyqtgraph as pg

import acconeer.exptool as et
from acconeer.exptool import a121
from acconeer.exptool.a121._core.entities.configs.sensor_config import SensorConfig
from acconeer.exptool.a121.algo.presence import Detector, DetectorConfig


def main():
    args = a121.ExampleArgumentParser().parse_args()
    et.utils.config_logging(args)

    client = a121.Client(**a121.get_client_args(args))
    client.connect()

    detector_config = DetectorConfig(
        start_m=1.0,
        end_m=3.0,
    )

    detector = Detector(client=client, sensor_id=1, detector_config=detector_config)
    detector.start()

    pg_updater = PGUpdater(detector_config, detector._get_sensor_config(detector_config))
    pg_process = et.PGProcess(pg_updater)
    pg_process.start()

    interrupt_handler = et.utils.ExampleInterruptHandler()
    print("Press Ctrl-C to end session")

    while not interrupt_handler.got_signal:
        detector_result = detector.get_next()
        print(detector_result)
        try:
            pg_process.put_data(detector_result)
        except et.PGProccessDiedException:
            break

    detector.stop()

    print("Disconnecting...")
    client.disconnect()


class PGUpdater:
    def __init__(self, detector_config: DetectorConfig, sensor_config: SensorConfig):
        self.detector_config = detector_config
        self.distances = np.linspace(
            detector_config.start_m, detector_config.end_m, sensor_config.num_points
        )

        self.history_length_s = 10

        max_num_of_sectors = max(6, self.distances.size // 3)
        self.sector_size = max(1, -(-self.distances.size // max_num_of_sectors))
        self.num_sectors = -(-self.distances.size // self.sector_size)
        self.sector_offset = (self.num_sectors * self.sector_size - self.distances.size) // 2

        self.setup_is_done = False

    def setup(self, win):
        win.setWindowTitle("Acconeer presence detection example")

        self.limit_lines = []
        dashed_pen = pg.mkPen("k", width=2.5)  # , style=QtCore.Qt.DashLine

        # Amplitude plot

        self.ampl_plot = win.addPlot(
            row=0,
            col=0,
            title="Amplitude, sweeps (orange), mean sweep (blue)",
        )

        self.ampl_plot.setMenuEnabled(False)
        self.ampl_plot.setMouseEnabled(x=False, y=False)
        self.ampl_plot.hideButtons()
        self.ampl_plot.showGrid(x=True, y=True)
        self.ampl_plot.setLabel("bottom", "Distance (m)")
        self.ampl_plot.setLabel("left", "Amplitude")

        self.ampl_plot.setYRange(0, 2**16)

        self.frame_scatter = pg.ScatterPlotItem(
            size=10,
            brush=et.utils.pg_brush_cycler(1),
        )
        self.mean_sweep_scatter = pg.ScatterPlotItem(
            size=10,
            brush=et.utils.pg_brush_cycler(0),
        )

        self.ampl_plot.addItem(self.frame_scatter)
        self.ampl_plot.addItem(self.mean_sweep_scatter)
        self.frame_smooth_limits = et.utils.SmoothLimits(self.detector_config.frame_rate)

        # Noise estimation plot

        self.noise_plot = win.addPlot(
            row=1,
            col=0,
            title="Noise",
        )
        self.noise_plot.setMenuEnabled(False)
        self.noise_plot.setMouseEnabled(x=False, y=False)
        self.noise_plot.hideButtons()
        self.noise_plot.showGrid(x=True, y=True)
        self.noise_plot.setLabel("bottom", "Distance (m)")
        self.noise_plot.setLabel("left", "Amplitude")
        self.noise_plot.setVisible(False)
        self.noise_curve = self.noise_plot.plot(pen=et.utils.pg_pen_cycler())
        self.noise_smooth_max = et.utils.SmoothMax(self.detector_config.frame_rate)

        # Depthwise presence plot

        self.move_plot = win.addPlot(
            row=2,
            col=0,
            title="Depthwise presence",
        )
        self.move_plot.setMenuEnabled(False)
        self.move_plot.setMouseEnabled(x=False, y=False)
        self.move_plot.hideButtons()
        self.move_plot.showGrid(x=True, y=True)
        self.move_plot.setLabel("bottom", "Distance (m)")
        self.move_plot.setLabel("left", "Norm. ampl.")
        zero_curve = self.move_plot.plot(self.distances, np.zeros_like(self.distances))
        self.inter_curve = self.move_plot.plot()
        self.total_curve = self.move_plot.plot()
        self.move_smooth_max = et.utils.SmoothMax(
            self.detector_config.frame_rate,
            tau_decay=1.0,
            tau_grow=0.25,
        )

        self.move_depth_line = pg.InfiniteLine(pen=pg.mkPen("k", width=1.5))
        self.move_depth_line.hide()
        self.move_plot.addItem(self.move_depth_line)
        limit_line = pg.InfiniteLine(angle=0, pen=dashed_pen)
        self.move_plot.addItem(limit_line)
        self.limit_lines.append(limit_line)

        fbi = pg.FillBetweenItem(
            zero_curve,
            self.inter_curve,
            brush=et.utils.pg_brush_cycler(0),
        )
        self.move_plot.addItem(fbi)

        fbi = pg.FillBetweenItem(
            self.inter_curve,
            self.total_curve,
            brush=et.utils.pg_brush_cycler(1),
        )
        self.move_plot.addItem(fbi)

        # Presence history plot

        self.move_hist_plot = pg.PlotItem(title="Presence history")
        self.move_hist_plot.setMenuEnabled(False)
        self.move_hist_plot.setMouseEnabled(x=False, y=False)
        self.move_hist_plot.hideButtons()
        self.move_hist_plot.showGrid(x=True, y=True)
        self.move_hist_plot.setLabel("bottom", "Time (s)")
        self.move_hist_plot.setLabel("left", "Score")
        self.move_hist_plot.setXRange(-self.history_length_s, 0)
        self.history_smooth_max = et.utils.SmoothMax(self.detector_config.frame_rate)
        self.move_hist_plot.setYRange(0, 10)

        self.move_hist_curve = self.move_hist_plot.plot(pen=et.utils.pg_pen_cycler())
        limit_line = pg.InfiniteLine(angle=0, pen=dashed_pen)
        self.move_hist_plot.addItem(limit_line)
        self.limit_lines.append(limit_line)

        self.present_html_format = (
            '<div style="text-align: center">'
            '<span style="color: #FFFFFF;font-size:15pt;">'
            "{}</span></div>"
        )
        not_present_html = (
            '<div style="text-align: center">'
            '<span style="color: #FFFFFF;font-size:15pt;">'
            "{}</span></div>".format("No presence detected")
        )
        self.present_text_item = pg.TextItem(
            fill=pg.mkColor(0xFF, 0x7F, 0x0E, 200),
            anchor=(0.5, 0),
        )
        self.not_present_text_item = pg.TextItem(
            html=not_present_html,
            fill=pg.mkColor(0x1F, 0x77, 0xB4, 180),
            anchor=(0.5, 0),
        )

        self.move_hist_plot.addItem(self.present_text_item)
        self.move_hist_plot.addItem(self.not_present_text_item)
        self.present_text_item.hide()
        self.not_present_text_item.hide()

        for line in self.limit_lines:
            line.setPos(self.detector_config.detection_threshold)

        # Sector plot

        self.sector_plot = pg.PlotItem()
        self.sector_plot.setAspectLocked()
        self.sector_plot.hideAxis("left")
        self.sector_plot.hideAxis("bottom")
        self.sectors = []

        pen = pg.mkPen("k", width=1)
        span_deg = 25
        for r in np.flip(np.arange(self.num_sectors) + 1):
            sector = pg.QtGui.QGraphicsEllipseItem(-r, -r, r * 2, r * 2)
            sector.setStartAngle(-16 * span_deg)
            sector.setSpanAngle(16 * span_deg * 2)
            sector.setPen(pen)
            self.sector_plot.addItem(sector)
            self.sectors.append(sector)

        self.sectors.reverse()

        sublayout = win.addLayout(row=3, col=0)
        sublayout.layout.setColumnStretchFactor(0, 2)
        sublayout.addItem(self.move_hist_plot, col=0)
        sublayout.addItem(self.sector_plot, col=1)

        self.setup_is_done = True

    def update(self, data):
        amplitudes = np.abs(data.processor_extra_result.frame)
        self.frame_scatter.setData(
            np.tile(self.distances, self.detector_config.sweeps_per_frame),
            amplitudes.flatten(),
        )

        self.mean_sweep_scatter.setData(self.distances, data.processor_extra_result.mean_sweep)
        self.ampl_plot.setYRange(*self.frame_smooth_limits.update(amplitudes))

        noise = data.processor_extra_result.lp_noise
        self.noise_curve.setData(self.distances, noise)
        self.noise_plot.setYRange(0, self.noise_smooth_max.update(noise))

        movement_x = data.presence_distance

        move_ys = data.processor_extra_result.depthwise_presence
        self.inter_curve.setData(self.distances, data.processor_extra_result.inter)
        self.total_curve.setData(self.distances, move_ys)
        m = self.move_smooth_max.update(np.max(move_ys))
        m = max(m, 2 * self.detector_config.detection_threshold)
        self.move_plot.setYRange(0, m)
        self.move_depth_line.setPos(movement_x)
        self.move_depth_line.setVisible(bool(data.presence_detected))

        move_hist_ys = data.processor_extra_result.presence_history
        move_hist_xs = np.linspace(-self.history_length_s, 0, len(move_hist_ys))

        m_hist = max(np.max(move_hist_ys), self.detector_config.detection_threshold * 1.05)
        m_hist = self.history_smooth_max.update(m_hist)

        self.move_hist_plot.setYRange(0, m_hist)
        self.move_hist_curve.setData(move_hist_xs, move_hist_ys)
        self.set_present_text_y_pos(m_hist)

        if data.presence_detected:
            present_text = "Presence detected at {:.0f} cm".format(movement_x * 100)
            present_html = self.present_html_format.format(present_text)
            self.present_text_item.setHtml(present_html)

            self.present_text_item.show()
            self.not_present_text_item.hide()
        else:
            self.present_text_item.hide()
            self.not_present_text_item.show()

        brush = et.utils.pg_brush_cycler(0)
        for sector in self.sectors:
            sector.setBrush(brush)

        if data.presence_detected:
            index = (
                data.processor_extra_result.presence_distance_index + self.sector_offset
            ) // self.sector_size
            self.sectors[index].setBrush(et.utils.pg_brush_cycler(1))

    def set_present_text_y_pos(self, y):
        self.present_text_item.setPos(-self.history_length_s / 2, 0.95 * y)
        self.not_present_text_item.setPos(-self.history_length_s / 2, 0.95 * y)


if __name__ == "__main__":
    main()



# Copyright (c) Acconeer AB, 2022
# All rights reserved

import numpy as np

from PySide6 import QtCore

import pyqtgraph as pg

import acconeer.exptool as et


class PGUpdater:
    def __init__(self, sensor_config, processing_config, session_info):
        self.sensor_config = sensor_config
        self.processing_config = processing_config

        self.history_length_s = processing_config.history_length_s
        self.depths = et.a111.get_range_depths(sensor_config, session_info)

        max_num_of_sectors = max(6, self.depths.size // 3)
        self.sector_size = max(1, -(-self.depths.size // max_num_of_sectors))
        self.num_sectors = -(-self.depths.size // self.sector_size)
        self.sector_offset = (self.num_sectors * self.sector_size - self.depths.size) // 2

        self.setup_is_done = False

        self.show_detect = 0

    def setup(self, win):
        win.setWindowTitle("Acconeer swipe algorithm example")

        self.limit_lines = []
        dashed_pen = pg.mkPen("k", width=2.5, style=QtCore.Qt.DashLine)

        # Data plot

        self.data_plot = win.addPlot(
            row=0,
            col=0,
            title="Frame movement",
        )
        self.data_plot.setMenuEnabled(False)
        self.data_plot.setMouseEnabled(x=False, y=False)
        self.data_plot.hideButtons()
        self.data_plot.showGrid(x=True, y=True)
        self.data_plot.setLabel("bottom", "Depth (m)")
        self.data_plot.setLabel("left", "Amplitude")
        self.data_plot.setYRange(0, 2**16)
        self.frame_scatter = pg.ScatterPlotItem(
            size=10,
            brush=et.utils.pg_brush_cycler(0),
        )

        self.data_plot.addItem(self.frame_scatter)
        self.frame_smooth_limits = et.utils.SmoothLimits(self.sensor_config.update_rate)

        # Trigger history plot

        self.move_hist_plot = pg.PlotItem(title="Presence history")
        self.move_hist_plot.setMenuEnabled(False)
        self.move_hist_plot.setMouseEnabled(x=False, y=False)
        self.move_hist_plot.hideButtons()
        self.move_hist_plot.showGrid(x=True, y=True)
        self.move_hist_plot.setLabel("bottom", "Time (s)")
        self.move_hist_plot.setLabel("left", "Score")
        self.move_hist_plot.setXRange(-self.history_length_s, 0)
        self.history_smooth_max = et.utils.SmoothMax(self.sensor_config.update_rate)
        self.move_hist_plot.setYRange(0, 10)

        self.move_hist_curve = self.move_hist_plot.plot(pen=et.utils.pg_pen_cycler())
        self.trig_line = pg.InfiniteLine(angle=0, pen=dashed_pen)
        self.cool_line = pg.InfiniteLine(angle=0, pen=dashed_pen)

        self.move_hist_plot.addItem(self.trig_line)
        self.move_hist_plot.addItem(self.cool_line)

        self.present_html_format = (
            '<div style="text-align: center">'
            '<span style="color: #FFFFFF;font-size:15pt;">'
            "{}</span></div>"
        )
        not_present_html = (
            '<div style="text-align: center">'
            '<span style="color: #FFFFFF;font-size:15pt;">'
            "{}</span></div>".format("No presence detected")
        )
        self.present_text_item = pg.TextItem(
            fill=pg.mkColor(0xFF, 0x7F, 0x0E, 200),
            anchor=(0.5, 0),
        )
        self.not_present_text_item = pg.TextItem(
            html=not_present_html,
            fill=pg.mkColor(0x1F, 0x77, 0xB4, 180),
            anchor=(0.5, 0),
        )

        self.move_hist_plot.addItem(self.present_text_item)
        self.move_hist_plot.addItem(self.not_present_text_item)
        self.present_text_item.hide()
        self.not_present_text_item.hide()

        sublayout = win.addLayout(row=3, col=0)
        sublayout.layout.setColumnStretchFactor(0, 2)
        sublayout.addItem(self.move_hist_plot, col=0)

        self.setup_is_done = True
        self.update_processing_config()

    def update_processing_config(self, processing_config=None):
        if processing_config is None:
            processing_config = self.processing_config
        else:
            self.processing_config = processing_config

        if not self.setup_is_done:
            return

        self.data_plot.setVisible(True)

        self.trig_line.setPos(processing_config.detection_threshold)
        self.cool_line.setPos(processing_config.cool_down_threshold)

    def update(self, data):
        self.frame_scatter.setData(
            np.tile(self.depths, self.sensor_config.sweeps_per_frame),
            data["frame"].flatten(),
        )

        self.data_plot.setYRange(*self.frame_smooth_limits.update(data["frame"]))

        move_hist_ys = data["presence_history"]
        move_hist_xs = np.linspace(-self.history_length_s, 0, len(move_hist_ys))

        m_hist = max(np.max(move_hist_ys), self.processing_config.detection_threshold * 1.05)
        m_hist = self.history_smooth_max.update(m_hist)

        if self.processing_config.history_plot_ceiling is not None:
            self.move_hist_plot.setYRange(0, self.processing_config.history_plot_ceiling)
            self.move_hist_curve.setData(
                move_hist_xs,
                np.minimum(move_hist_ys, self.processing_config.history_plot_ceiling),
            )
            self.set_present_text_y_pos(self.processing_config.history_plot_ceiling)
        else:
            self.move_hist_plot.setYRange(0, m_hist)
            self.move_hist_curve.setData(move_hist_xs, move_hist_ys)
            self.set_present_text_y_pos(m_hist)

        if any(data["detection_history"][-5:]):
            self.show_detect = 15

        if self.show_detect > 0:
            present_text = "Wave!"
            present_html = self.present_html_format.format(present_text)
            self.present_text_item.setHtml(present_html)

            self.present_text_item.show()
            self.show_detect -= 1
        else:
            self.present_text_item.hide()

    def set_present_text_y_pos(self, y):
        self.present_text_item.setPos(-self.history_length_s / 2, 0.95 * y)
        self.not_present_text_item.setPos(-self.history_length_s / 2, 0.95 * y)



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from __future__ import annotations

import datetime
import os
from pathlib import Path
from typing import Any, Optional, TypeVar
from uuid import uuid4

import h5py
import importlib_metadata
import numpy as np

from acconeer.exptool.a121._core.entities import (
    INT_16_COMPLEX,
    ClientInfo,
    Metadata,
    Result,
    ServerInfo,
    SessionConfig,
)
from acconeer.exptool.a121._core.mediators import Recorder

from .utils import PathOrH5File, h5_file_factory


T = TypeVar("T")


def get_h5py_str_dtype():
    return h5py.special_dtype(vlen=str)


_H5PY_STR_DTYPE = get_h5py_str_dtype()


def get_timestamp() -> str:
    return datetime.datetime.now().isoformat(timespec="seconds")


def get_uuid() -> str:
    return str(uuid4())


class H5Recorder(Recorder):
    path: Optional[os.PathLike]
    file: h5py.File
    owns_file: bool
    _num_frames: int

    def __init__(
        self,
        path_or_file: PathOrH5File,
        mode: str = "x",
        *,
        _lib_version: Optional[str] = None,
        _timestamp: Optional[str] = None,
        _uuid: Optional[str] = None,
    ) -> None:
        self.file, self.owns_file = h5_file_factory(path_or_file, h5_file_mode=mode)
        self.path = Path(self.file.filename) if self.owns_file else None

        if _lib_version is None:
            _lib_version = importlib_metadata.version("acconeer-exptool")

        if _timestamp is None:
            _timestamp = get_timestamp()

        if _uuid is None:
            _uuid = get_uuid()

        self._num_frames = 0

        self.file.create_dataset(
            "uuid",
            data=_uuid,
            dtype=_H5PY_STR_DTYPE,
            track_times=False,
        )
        self.file.create_dataset(
            "timestamp",
            data=_timestamp,
            dtype=_H5PY_STR_DTYPE,
            track_times=False,
        )
        self.file.create_dataset(
            "lib_version",
            data=_lib_version,
            dtype=_H5PY_STR_DTYPE,
            track_times=False,
        )
        self.file.create_dataset(
            "generation",
            data="a121",
            dtype=_H5PY_STR_DTYPE,
            track_times=False,
        )

    def _start(
        self,
        *,
        client_info: ClientInfo,
        extended_metadata: list[dict[int, Metadata]],
        server_info: ServerInfo,
        session_config: SessionConfig,
    ) -> None:
        self.file.create_dataset(
            "client_info",
            data=client_info.to_json(),
            dtype=_H5PY_STR_DTYPE,
            track_times=False,
        )
        self.file.create_dataset(
            "server_info",
            data=server_info.to_json(),
            dtype=_H5PY_STR_DTYPE,
            track_times=False,
        )

        session_group = self.file.create_group("session")

        session_group.create_dataset(
            "session_config",
            data=session_config.to_json(),
            dtype=_H5PY_STR_DTYPE,
            track_times=False,
        )

        for i, metadata_group_dict in enumerate(extended_metadata):
            group_group = session_group.create_group(f"group_{i}")

            for entry_id, (sensor_id, metadata) in enumerate(metadata_group_dict.items()):
                entry_group = group_group.create_group(f"entry_{entry_id}")
                entry_group.create_dataset("sensor_id", data=sensor_id, track_times=False)

                entry_group.create_dataset(
                    "metadata",
                    data=metadata.to_json(),
                    dtype=_H5PY_STR_DTYPE,
                    track_times=False,
                )

                result_group = entry_group.create_group("result")
                self._create_result_datasets(result_group, metadata)

    def _sample(self, extended_result: list[dict[int, Result]]) -> None:
        for group_index, result_group_dict in enumerate(extended_result):
            for entry_id, result in enumerate(result_group_dict.values()):
                result_group = self.file[f"session/group_{group_index}/entry_{entry_id}/result"]
                self._write_result(result_group, self._num_frames, result)

        self._num_frames += 1

    def _stop(self) -> Any:
        if self.owns_file:
            self.file.close()

    @staticmethod
    def _create_result_datasets(g: h5py.Group, metadata: Metadata) -> None:
        g.create_dataset(
            "data_saturated",
            shape=(0,),
            maxshape=(None,),
            dtype=bool,
            track_times=False,
        )
        g.create_dataset(
            "frame_delayed",
            shape=(0,),
            maxshape=(None,),
            dtype=bool,
            track_times=False,
        )
        g.create_dataset(
            "calibration_needed",
            shape=(0,),
            maxshape=(None,),
            dtype=bool,
            track_times=False,
        )
        g.create_dataset(
            "temperature",
            shape=(0,),
            maxshape=(None,),
            dtype=int,
            track_times=False,
        )

        g.create_dataset(
            "tick",
            shape=(0,),
            maxshape=(None,),
            dtype=np.dtype("int64"),
            track_times=False,
        )

        g.create_dataset(
            "frame",
            shape=(0, *metadata.frame_shape),
            maxshape=(None, *metadata.frame_shape),
            dtype=INT_16_COMPLEX,
            track_times=False,
        )

    @staticmethod
    def _write_result(g: h5py.Group, index: int, result: Result) -> None:
        """Extends the Dataset to the appropriate (new) size with .resize,
        and then copies the data over
        """
        datasets_to_extend = [
            "data_saturated",
            "frame_delayed",
            "calibration_needed",
            "temperature",
            "tick",
            "frame",
        ]
        for dataset_name in datasets_to_extend:
            g[dataset_name].resize(size=index + 1, axis=0)

        g["data_saturated"][index] = result.data_saturated
        g["frame_delayed"][index] = result.frame_delayed
        g["calibration_needed"][index] = result.calibration_needed
        g["temperature"][index] = result.temperature
        g["tick"][index] = result.tick
        g["frame"][index] = result._frame

    def require_algo_group(self, key: str) -> h5py.Group:
        group = self.file.require_group("algo")

        if "key" in group:
            existing_key = bytes(group["key"][()]).decode()
            if existing_key != key:
                raise Exception(f"Algo group key mismatch: got '{key}' but had '{existing_key}'")
        else:
            group.create_dataset(
                "key",
                data=key,
                dtype=_H5PY_STR_DTYPE,
                track_times=False,
            )

        return group



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from __future__ import annotations

import enum
from typing import Optional

import attrs
import numpy as np
import numpy.typing as npt

from acconeer.exptool import a121
from acconeer.exptool.a121 import Profile
from acconeer.exptool.a121.algo import AlgoConfigBase, AlgoParamEnum, ProcessorBase


class MeasurementType(AlgoParamEnum):
    CLOSE_RANGE = enum.auto()
    FAR_RANGE = enum.auto()
    CLOSE_AND_FAR_RANGE = enum.auto()


@attrs.mutable(kw_only=True)
class ProcessorConfig(AlgoConfigBase):

    measurement_type: MeasurementType = attrs.field(
        default=MeasurementType.CLOSE_RANGE,
        converter=MeasurementType,
    )
    """The measurement type.
    ``CLOSE_RANGE`` corresponds to a range of approximate 2 cm - 12 cm.
    ``FAR_RANGE`` corresponds to approximate 10 cm - 30 cm.
    ``CLOSE_AND_FAR_RANGE`` gives two detection outputs, one for each range."""

    sensitivity_close: float = attrs.field(default=2.5)
    """Sensitivity for close range detection. High sensitivity equals low detection threshold,
    low sensitivity equals high detection threshold."""

    sensitivity_far: float = attrs.field(default=1.4)
    """Sensitivity for far range detection. High sensitivity equals low detection threshold,
    low sensitivity equals high detection threshold."""

    patience_close: int = attrs.field(default=2)
    """Number of frames in a row above threshold to count as a new close range detection,
    also number of frames in a row below threshold to count as end of detection."""

    patience_far: int = attrs.field(default=2)
    """Number of frames in a row above threshold to count as a new far range detection,
    also number of frames in a row below threshold to count as end of detection."""

    calibration_duration_s: float = attrs.field(default=0.6)
    """Calibration duration in seconds"""

    calibration_interval_s: float = attrs.field(default=20)
    """Interval between calibrations in seconds. When reached a new calibration is made.
    Should not be set lower than the longest estimated continuous detection event."""


@attrs.frozen(kw_only=True)
class ProcessorResult:
    detection_close: Optional[bool] = attrs.field(default=None)
    detection_far: Optional[bool] = attrs.field(default=None)


class Processor(ProcessorBase[ProcessorConfig, ProcessorResult]):
    """Virtual Button processor

    :param sensor_config: Sensor configuration
    :param metadata: Metadata yielded by the sensor config
    :param processor_config: Processor configuration
    """

    def __init__(
        self,
        *,
        sensor_config: a121.SensorConfig,
        metadata: a121.Metadata,
        processor_config: ProcessorConfig,
    ) -> None:

        self._sensor_config = sensor_config
        self._metadata = metadata
        self._processor_config = processor_config

        self._sweeps_per_frame = sensor_config.sweeps_per_frame

        if sensor_config.sweep_rate is None:
            raise ValueError("sweep_rate must be set")

        frame_rate = sensor_config.sweep_rate / sensor_config.sweeps_per_frame
        self._cal_interval_frames = int(processor_config.calibration_interval_s * frame_rate)

        self._calibrate()

        self._detection_close = False
        self._detection_far = False

        self._sig_count = np.zeros((2,))
        self._nonsig_count = np.array(
            [processor_config.patience_close + 1, processor_config.patience_far + 1],
        )

    def process(self, result: a121.Result) -> ProcessorResult:
        frame = result.frame
        y = np.zeros_like(frame, dtype=float)

        sensitivity = self._get_sensitivity()

        if self._frames_since_last_cal > self._cal_interval_frames:
            self._calibrate()
        elif not np.any(np.isnan(self._cfar_ref_buf)):
            y = self._calc_variance(frame)

        significant = y > 1 / sensitivity * 10
        detection_depth, counts = np.unique(np.where(significant)[1], return_counts=True)

        if self._processor_config.measurement_type == MeasurementType.CLOSE_RANGE:
            self._detection_close = self._process_single_range(
                frame=frame,
                detection=self._detection_close,
                patience=self._processor_config.patience_close,
                detection_depth=detection_depth,
                counts=counts,
                measurement_type=MeasurementType.CLOSE_RANGE,
            )
            return ProcessorResult(detection_close=self._detection_close)

        if self._processor_config.measurement_type == MeasurementType.FAR_RANGE:
            self._detection_far = self._process_single_range(
                frame=frame,
                detection=self._detection_far,
                patience=self._processor_config.patience_far,
                detection_depth=detection_depth,
                counts=counts,
                measurement_type=MeasurementType.FAR_RANGE,
            )
            return ProcessorResult(detection_far=self._detection_far)

        if self._processor_config.measurement_type == MeasurementType.CLOSE_AND_FAR_RANGE:
            return self._process_multiple_ranges(
                frame=frame,
                detection_depth=detection_depth,
                counts=counts,
            )

        raise AssertionError

    def _calibrate(self) -> None:
        assert self._sensor_config.sweep_rate is not None
        self._cfar_ref_buf = np.full(
            (
                int(
                    self._processor_config.calibration_duration_s * self._sensor_config.sweep_rate
                ),
                self._metadata.sweep_data_length,
            ),
            np.nan,
            dtype="complex",
        )
        self._cfar_guard_buf = np.full(
            (self._sweeps_per_frame, self._metadata.sweep_data_length),
            np.nan,
            dtype="complex",
        )
        self._frames_since_last_cal = 0

    def _calc_variance(self, frame: npt.NDArray) -> npt.NDArray:
        xn = np.full((self._sweeps_per_frame, self._metadata.sweep_data_length), 0, dtype=float)
        y = np.full((self._sweeps_per_frame, self._metadata.sweep_data_length), 0, dtype=float)

        arg_norm = np.mean(self._cfar_ref_buf, axis=0)
        arg_norm = np.conj(arg_norm) / np.abs(arg_norm)

        arg_norm_ref = self._cfar_ref_buf * arg_norm
        ref_ampls = np.abs(arg_norm_ref)
        ampl_mean = np.mean(ref_ampls, axis=0)
        ampl_std = np.std(ref_ampls, axis=0)
        ref_phases = np.angle(arg_norm_ref)
        phase_mean = np.mean(ref_phases, axis=0)
        phase_std = np.std(ref_phases, axis=0)

        xn = frame * arg_norm

        y = np.hypot(
            (np.abs(xn) - ampl_mean) / ampl_std,
            (np.angle(xn) - phase_mean) / phase_std,
        )

        return y

    def _process_single_range(
        self,
        frame: npt.NDArray,
        detection: bool,
        patience: int,
        detection_depth: npt.NDArray,
        counts: npt.NDArray,
        measurement_type: MeasurementType,
    ) -> bool:

        if measurement_type == MeasurementType.CLOSE_RANGE:
            index = 0
        elif measurement_type == MeasurementType.FAR_RANGE:
            index = 1
        else:
            raise AssertionError

        # Checks if there are at least two significant values at the same depth
        if detection_depth[counts > 1].size > 0:
            self._sig_count[index] += 1
            self._nonsig_count[index] = 0
            self._cfar_guard_buf = np.full(
                (
                    self._sensor_config.sweeps_per_frame,
                    self._metadata.sweep_data_length,
                ),
                np.nan,
                dtype="complex",
            )
            self._frames_since_last_cal += 1
        else:
            self._sig_count[index] = 0
            self._nonsig_count[index] += 1

            if not np.isnan(self._cfar_guard_buf).any():
                self._update_cfar_buffer()

            self._cfar_guard_buf = frame

        detection = self._get_detection(
            detection,
            self._sig_count[index],
            self._nonsig_count[index],
            patience,
        )

        return detection

    def _process_multiple_ranges(
        self,
        frame: npt.NDArray,
        detection_depth: npt.NDArray,
        counts: npt.NDArray,
    ) -> ProcessorResult:

        if detection_depth[counts > 1].size > 0:
            if (
                np.where(
                    detection_depth[counts > 1] < self._sensor_config.subsweeps[0].num_points,
                    True,
                    False,
                )
            ).any():
                self._sig_count[0] += 1
                self._nonsig_count[0] = 0
            else:
                self._sig_count[0] = 0
                self._nonsig_count[0] += 1
            if (
                np.where(
                    detection_depth[counts > 1] >= self._sensor_config.subsweeps[0].num_points,
                    True,
                    False,
                )
            ).any():
                self._sig_count[1] += 1
                self._nonsig_count[1] = 0
            else:
                self._sig_count[1] = 0
                self._nonsig_count[1] += 1

            self._cfar_guard_buf = np.full(
                (
                    self._sensor_config.sweeps_per_frame,
                    self._metadata.sweep_data_length,
                ),
                np.nan,
                dtype="complex",
            )
            self._frames_since_last_cal += 1

        else:
            self._sig_count[0] = 0
            self._nonsig_count[0] += 1
            self._sig_count[1] = 0
            self._nonsig_count[1] += 1

            if not np.isnan(self._cfar_guard_buf).any():
                self._update_cfar_buffer()

            self._cfar_guard_buf = frame

        self._detection_close = self._get_detection(
            self._detection_close,
            self._sig_count[0],
            self._nonsig_count[0],
            self._processor_config.patience_close,
        )

        self._detection_far = self._get_detection(
            self._detection_far,
            self._sig_count[1],
            self._nonsig_count[1],
            self._processor_config.patience_far,
        )

        return ProcessorResult(
            detection_close=self._detection_close,
            detection_far=self._detection_far,
        )

    def _get_sensitivity(self) -> npt.NDArray[np.float_]:
        if self._processor_config.measurement_type == MeasurementType.FAR_RANGE:
            return np.repeat(
                self._processor_config.sensitivity_far,
                self._sensor_config.num_points,
            )
        elif self._processor_config.measurement_type == MeasurementType.CLOSE_RANGE:
            return np.repeat(
                self._processor_config.sensitivity_close,
                self._sensor_config.num_points,
            )
        elif self._processor_config.measurement_type == MeasurementType.CLOSE_AND_FAR_RANGE:
            return np.repeat(
                [
                    self._processor_config.sensitivity_close,
                    self._processor_config.sensitivity_far,
                ],
                [
                    self._sensor_config.subsweeps[0].num_points,
                    self._sensor_config.subsweeps[1].num_points,
                ],
            )

        raise AssertionError

    def _update_cfar_buffer(self) -> None:
        self._cfar_ref_buf = np.roll(self._cfar_ref_buf, -self._sweeps_per_frame, axis=0)
        self._cfar_ref_buf[-self._sweeps_per_frame - 1 : -1, :] = self._cfar_guard_buf
        self._frames_since_last_cal = 0

    @staticmethod
    def _get_detection(
        curr_detection: object, sig_count: int, nonsig_count: int, patience: int
    ) -> bool:
        new_detection = not curr_detection and sig_count >= patience
        keep_detection = curr_detection and nonsig_count <= patience
        detection = new_detection or keep_detection
        return bool(detection)

    def update_config(self, config: ProcessorConfig) -> None:
        raise NotImplementedError


def get_sensor_config() -> a121.SensorConfig:
    return a121.SensorConfig(
        sweeps_per_frame=16,
        sweep_rate=320,
        continuous_sweep_mode=True,
        double_buffering=True,
        inter_sweep_idle_state=a121.IdleState.READY,
        inter_frame_idle_state=a121.IdleState.READY,
        num_points=4,
        profile=Profile.PROFILE_1,
        receiver_gain=0,
        hwaas=20,
        start_point=14,
        step_length=12,
    )



# Copyright (c) Acconeer AB, 2022
# All rights reserved

import acconeer.exptool as et
from acconeer.exptool import a121


args = a121.ExampleArgumentParser().parse_args()
et.utils.config_logging(args)

client = a121.Client(**a121.get_client_args(args))
client.connect()

sensor_config = a121.SensorConfig(
    subsweeps=[
        a121.SubsweepConfig(
            start_point=25,
            step_length=2,
            num_points=30,
            profile=a121.Profile.PROFILE_1,
            hwaas=10,
        ),
        a121.SubsweepConfig(
            start_point=75,
            step_length=4,
            num_points=25,
            profile=a121.Profile.PROFILE_3,
            hwaas=20,
        ),
    ],
    sweeps_per_frame=10,
)

client.setup_session(sensor_config)
client.start_session()

for i in range(3):
    result = client.get_next()

    print(f"\nResult {i + 1} subframes:")
    print(result.subframes)

client.stop_session()
client.disconnect()



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from __future__ import annotations

from typing import Optional, Tuple

import numpy as np


class RateCalculator:
    def __init__(self) -> None:
        self._reset()

    def _reset(self):
        self.last_time = None
        self.fifo = np.full(200, np.nan)

    def update(self, time: Optional[float]) -> Tuple[float, float]:
        if time is None:
            self._reset()
            return np.nan, np.nan

        last_time = self.last_time
        self.last_time = time

        if last_time is None:
            return np.nan, np.nan

        delta_time = time - last_time

        self.fifo = np.roll(self.fifo, -1)
        self.fifo[-1] = delta_time

        rate = 1.0 / np.nanmean(self.fifo)
        jitter = np.nanstd(self.fifo)

        return float(rate), float(jitter)



# Copyright (c) Acconeer AB, 2022
# All rights reserved

import acconeer.exptool as et


def main():
    args = et.a111.ExampleArgumentParser().parse_args()
    et.utils.config_logging(args)

    client = et.a111.Client(**et.a111.get_client_args(args))

    # Normally when using a single sensor, get_next will return
    # (info, data). When using mulitple sensors, get_next will return
    # lists of info and data for each sensor, i.e. ([info], [data]).
    # This is commonly called squeezing. To disable squeezing, making
    # get_next _always_ return lists, set:
    # client.squeeze = False

    config = et.a111.EnvelopeServiceConfig()
    config.sensor = args.sensors
    config.range_interval = [0.2, 0.3]
    config.update_rate = 5

    session_info = client.setup_session(config)
    print("Session info:\n", session_info, "\n")

    # Now would be the time to set up plotting, signal processing, etc.

    client.start_session()

    # Normally, hitting Ctrl-C will raise a KeyboardInterrupt which in
    # most cases immediately terminates the script. This often becomes
    # an issue when plotting and also doesn't allow us to disconnect
    # gracefully. Setting up an ExampleInterruptHandler will capture the
    # keyboard interrupt signal so that a KeyboardInterrupt isn't raised
    # and we can take care of the signal ourselves. In case you get
    # impatient, hitting Ctrl-C a couple of more times will raise a
    # KeyboardInterrupt which hopefully terminates the script.
    interrupt_handler = et.utils.ExampleInterruptHandler()
    print("Press Ctrl-C to end session\n")

    while not interrupt_handler.got_signal:
        info, data = client.get_next()
        print(info, "\n", data, "\n")

    print("Disconnecting...")
    client.disconnect()


if __name__ == "__main__":
    main()



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from __future__ import annotations

import enum
from typing import Optional, Tuple

import attrs
import numpy as np
import numpy.typing as npt

from acconeer.exptool import a121
from acconeer.exptool.a121.algo import AlgoParamEnum
from acconeer.exptool.a121.algo.distance._processors import (
    Processor,
    ProcessorConfig,
    ProcessorContext,
    ProcessorResult,
)


class PeakSortingMethod(AlgoParamEnum):
    """Peak sorting methods.
    ``CLOSEST`` sort according to distance.
    ``STRONGEST`` sort according to amplitude
    ``HIGHEST_RCS`` sort according to RCS."""

    CLOSEST = enum.auto()
    STRONGEST = enum.auto()
    HIGHEST_RCS = enum.auto()


@attrs.frozen(kw_only=True)
class ProcessorSpec:
    processor_config: ProcessorConfig = attrs.field()
    group_index: int = attrs.field()
    sensor_id: int = attrs.field()
    subsweep_indexes: list[int] = attrs.field()
    processor_context: Optional[ProcessorContext] = attrs.field(default=None)


@attrs.mutable(kw_only=True)
class AggregatorConfig:
    peak_sorting_method: PeakSortingMethod = attrs.field(default=PeakSortingMethod.STRONGEST)


@attrs.frozen(kw_only=True)
class AggregatorContext:
    offset_m: float = attrs.field(default=0.0)


@attrs.frozen(kw_only=True)
class AggregatorResult:
    processor_results: list[ProcessorResult] = attrs.field()
    estimated_distances: npt.NDArray[np.float_] = attrs.field()
    service_extended_result: list[dict[int, a121.Result]] = attrs.field()


class Aggregator:
    """Aggregating class

    Instantiates Processor objects based configuration from Detector.

    Aggregates result, based on selected peak sorting strategy, from underlying Processor objects.
    """

    MIN_PEAK_DIST_M = 0.005

    def __init__(
        self,
        session_config: a121.SessionConfig,
        extended_metadata: list[dict[int, a121.Metadata]],
        config: AggregatorConfig,
        context: AggregatorContext,
        specs: list[ProcessorSpec],
    ):
        self.config = config
        self.context = context
        self.specs = specs

        self.processors: list[Processor] = []

        for spec in specs:
            metadata = extended_metadata[spec.group_index][spec.sensor_id]
            sensor_config = session_config.groups[spec.group_index][spec.sensor_id]

            processor = Processor(
                sensor_config=sensor_config,
                metadata=metadata,
                processor_config=spec.processor_config,
                subsweep_indexes=spec.subsweep_indexes,
                context=spec.processor_context,
            )
            self.processors.append(processor)

    def process(self, extended_result: list[dict[int, a121.Result]]) -> AggregatorResult:
        processors_result = []
        ampls: npt.NDArray[np.float_] = np.array([])
        dists: npt.NDArray[np.float_] = np.array([])
        for spec, processor in zip(self.specs, self.processors):
            processor_result = processor.process(extended_result[spec.group_index][spec.sensor_id])
            processors_result.append(processor_result)
            if processor_result.estimated_distances is not None:
                ampls = np.concatenate((ampls, np.array(processor_result.estimated_amplitudes)))
                dists = np.concatenate((dists, np.array(processor_result.estimated_distances)))

        (dists_merged, ampls_merged) = self._merge_peaks(self.MIN_PEAK_DIST_M, dists, ampls)
        dists_sorted = self._sort_peaks(
            dists_merged, ampls_merged, self.config.peak_sorting_method
        )
        dists_sorted -= self.context.offset_m
        return AggregatorResult(
            processor_results=processors_result,
            estimated_distances=dists_sorted,
            service_extended_result=extended_result,
        )

    @staticmethod
    def _merge_peaks(
        min_peak_to_peak_dist: float,
        dists: npt.NDArray[np.float_],
        ampls: npt.NDArray[np.float_],
    ) -> Tuple[npt.NDArray[np.float_], npt.NDArray[np.float_]]:
        sorting_order = np.argsort(dists)
        distances_sorted = dists[sorting_order]
        amplitudes_sorted = ampls[sorting_order]

        peak_cluster_idxs = np.where(min_peak_to_peak_dist < np.diff(distances_sorted))[0] + 1
        distances_merged = [
            np.mean(cluster) for cluster in np.split(distances_sorted, peak_cluster_idxs)
        ]
        amplitudes_merged = [
            np.mean(cluster) for cluster in np.split(amplitudes_sorted, peak_cluster_idxs)
        ]
        return (np.array(distances_merged), np.array(amplitudes_merged))

    @staticmethod
    def _sort_peaks(
        dists: npt.NDArray[np.float_],
        ampls: npt.NDArray[np.float_],
        method: PeakSortingMethod,
    ) -> npt.NDArray[np.float_]:
        if method == PeakSortingMethod.CLOSEST:
            quantity_to_sort = dists
        elif method == PeakSortingMethod.STRONGEST:
            quantity_to_sort = -ampls
        elif method == PeakSortingMethod.HIGHEST_RCS:
            quantity_to_sort = -ampls * dists**2
        else:
            raise ValueError("Unknown peak sorting method")
        return np.array([dists[i] for i in quantity_to_sort.argsort()])
